import sys
import os
import onnx
from onnx import helper, TensorProto, numpy_helper
import unittest
import numpy as np

sys.path.append('..')
from neural_compressor.model.onnx_model import ONNXModel

def get_onnx_model():
    model = torchvision.models.resnet18()
    x = Variable(torch.randn(1, 3, 224, 224))
    torch_out = torch.onnx.export(model, x, "resnet18.onnx", export_params=True, verbose=True)

def generate_input_initializer(tensor_shape, tensor_dtype, input_name):
    '''
    Helper function to generate initializers for test inputs
    '''
    tensor = np.random.ranf(tensor_shape).astype(tensor_dtype)
    init = numpy_helper.from_array(tensor, input_name)
    return init  

class TestOnnxModelUtil(unittest.TestCase):
    def setUp(self):
        #   input
        #    |
        #   Mul
        #    |  \
        #    |    initializer
        #    |  /
        #   Mul
        input0 = helper.make_tensor_value_info('input0', TensorProto.FLOAT, [1, 3])
        output = helper.make_tensor_value_info('output', TensorProto.FLOAT, [1, 3])
        
        mul_weight = generate_input_initializer([], np.float32, 'mul_weight')
       
        mul_node_1 = onnx.helper.make_node('Mul', ['input0', 'mul_weight'], ['mul_1'], name='Mul1')
        mul_node_2 = onnx.helper.make_node('Mul', ['mul_1', 'mul_weight'], ['output'], name= 'Mul2')
        
        graph = helper.make_graph([mul_node_1, mul_node_2], 'test_graph', [input0], [output])
        graph.initializer.add().CopyFrom(mul_weight)

        model = helper.make_model(graph)
        test_model_path = './test_model.onnx'
        onnx.save(model, test_model_path)
        model = onnx.load(test_model_path)
        self.model = ONNXModel(model)

    def test_split_shared_input(self):
        from neural_compressor.adaptor.ox_utils.util import split_shared_input
        model = split_shared_input(self.model)
        self.assertIsNone(model.get_initializer('mul_weight_nc_split_Mul1'))
        self.assertIsNotNone(model.get_initializer('mul_weight_nc_split_Mul2'))

    def tearDown(self):
        os.remove('./test_model.onnx')

class TestOnnxModel(unittest.TestCase):
    def setUp(self):
        #   Relu      
        #    |      \ 
        #   Conv     \
        #    |        \ 
        #   Relu       |  
        #    |       Conv  
        #   Conv      / 
        #      \     /  
        #         |
        #        Add
    
        input0 = helper.make_tensor_value_info('input0', TensorProto.FLOAT, [1, 3, 1, 3])
        output = helper.make_tensor_value_info('output', TensorProto.FLOAT, [1, 3, 1, 3])
        
        X1_weight = generate_input_initializer([3, 3, 1, 1], np.float32, 'X1_weight')
        X1_bias = generate_input_initializer([3], np.float32, 'X1_bias')
        X3_weight = generate_input_initializer([3, 3, 1, 1], np.float32, 'X3_weight')
        X3_bias = generate_input_initializer([3],np.float32, 'X3_bias')
        X5_weight = generate_input_initializer([3, 3, 1, 1], np.float32, 'X5_weight')
        X5_bias = generate_input_initializer([3],np.float32,'X5_bias')
       
        relu_node_1 = onnx.helper.make_node('Relu', ['input0'], ['X1'], name='Relu1')
        conv_node_1 = onnx.helper.make_node('Conv', ['X1', 'X1_weight', 'X1_bias'], ['X2'], name='Conv1')
        relu_node_2 = onnx.helper.make_node('Relu', ['X2'], ['X3'], name= 'Relu2')
        conv_node_2 = onnx.helper.make_node('Conv', ['X3', 'X3_weight', 'X3_bias'], ['X4'], name='Conv2')
        conv_node_3 = onnx.helper.make_node('Conv', ['X1', 'X5_weight', 'X5_bias'], ['X5'], name='Conv3')
        add_node = onnx.helper.make_node('Add', ['X4', 'X5'], ['output'], name='Add')
      
        graph = helper.make_graph([relu_node_1, conv_node_1, relu_node_2, conv_node_2, conv_node_3, add_node], 'test_graph_6', [input0], [output])
        graph.initializer.add().CopyFrom(X1_weight)
        graph.initializer.add().CopyFrom(X1_bias)
        graph.initializer.add().CopyFrom(X3_weight)
        graph.initializer.add().CopyFrom(X3_bias)
        graph.initializer.add().CopyFrom(X5_weight)
        graph.initializer.add().CopyFrom(X5_bias)

        model = helper.make_model(graph)
        test_model_path = './test_model_6.onnx'
        onnx.save(model, test_model_path)
        model = onnx.load(test_model_path)
        self.model = ONNXModel(model)

        #    QuantizeLinear
        #        |
        #    QLinearConv
        #        |
        #    DequantizeLinear
        A = helper.make_tensor_value_info('A', TensorProto.FLOAT, [1, 1, 5, 5])
        A_scale = helper.make_tensor_value_info('A_scale', TensorProto.FLOAT, [1])
        a_scale = generate_input_initializer([1], np.float32, 'A_scale')
        A_zero = helper.make_tensor_value_info('A_zero_point', TensorProto.INT8, [1])
        a_zero_point = generate_input_initializer([1], np.int8, 'A_zero_point')
        B_scale = helper.make_tensor_value_info('B_scale', TensorProto.FLOAT, [1])
        b_scale = generate_input_initializer([1], np.float32, 'B_scale')
        B_zero = helper.make_tensor_value_info('B_zero_point', TensorProto.INT8, [1])
        b_zero_point = generate_input_initializer([1], np.int8, 'B_zero_point')
        C = helper.make_tensor_value_info('C', TensorProto.INT8, [1, 1, 5, 5])
        c = generate_input_initializer([1, 1, 5, 5], np.int8, 'C')
        C_scale = helper.make_tensor_value_info('C_scale', TensorProto.FLOAT, [1])
        c_scale = generate_input_initializer([1], np.float32, 'C_scale')
        C_zero = helper.make_tensor_value_info('C_zero_point', TensorProto.INT8, [1])
        c_zero_point = generate_input_initializer([1], np.int8, 'C_zero_point')
        E = helper.make_tensor_value_info('E', TensorProto.INT32, [1])
        e = generate_input_initializer([1], np.int32, 'E')
        D_scale = helper.make_tensor_value_info('D_scale', TensorProto.FLOAT, [1])
        d_scale = generate_input_initializer([1], np.float32, 'D_scale')
        D_zero = helper.make_tensor_value_info('D_zero_point', TensorProto.INT8, [1])
        d_zero_point = generate_input_initializer([1], np.int8, 'D_zero_point')
        D = helper.make_tensor_value_info('D', TensorProto.FLOAT, [1, 1, 5, 5])
        quantize_node = onnx.helper.make_node('QuantizeLinear', ['A', 'A_scale', 'A_zero_point'], ['B_quantized'], name='A_QuantizeLinear')
        conv_node = onnx.helper.make_node('QLinearConv', ['B_quantized', 'B_scale', 'B_zero_point', 'C_quantized', 'C_scale', 'C_zero_point', 'D_scale', 'D_zero_point', 'E'], ['D_quantized'], name='conv_quant', kernel_shape=[3, 3], pads=[1, 1, 1, 1])
        dequantize_node = onnx.helper.make_node('DequantizeLinear', ['D_quantized', 'D_scale', 'D_zero_point'], ['D'], name='D_DequantizeLinear')
        graph = helper.make_graph([quantize_node, conv_node, dequantize_node], 'test_graph_7', [A, A_scale, A_zero, C, C_scale, C_zero, E, D_scale, D_zero], [D])
        graph.initializer.add().CopyFrom(a_scale)
        graph.initializer.add().CopyFrom(a_zero_point)
        graph.initializer.add().CopyFrom(b_scale)
        graph.initializer.add().CopyFrom(b_zero_point)
        graph.initializer.add().CopyFrom(c)
        graph.initializer.add().CopyFrom(c_scale)
        graph.initializer.add().CopyFrom(c_zero_point)
        graph.initializer.add().CopyFrom(e)
        graph.initializer.add().CopyFrom(d_scale)
        graph.initializer.add().CopyFrom(d_zero_point)
        model = helper.make_model(graph)
        self.q_model = ONNXModel(model)

    def test_nodes(self):
        self.assertEqual(len(self.model.nodes()), 6)
        nodes_name = [node.name for node in self.model.nodes()]
        nodes = ["Relu1", "Conv1", "Relu2", "Conv2", "Conv3", "Add"]
        for node in nodes:
            self.assertTrue(node in nodes_name)

    def test_initializer(self):
        self.assertEqual(len(self.model.initializer()), 6)
        inits_name = [init.name for init in self.model.initializer()]
        inits = ['X1_weight', 'X1_bias', 'X3_weight', 'X3_bias', 'X5_weight', 'X5_bias']
        for init in inits:
            self.assertTrue(init in inits_name)
        

    def test_remove_node(self):
        for node in self.model.nodes():
            if node.op_type == "Add":
                self.model.remove_node(node)
        self.assertEqual(len(self.model.nodes()), 5)
        nodes_name = [node.name for node in self.model.nodes()]
        nodes = ["Relu1", "Conv1", "Relu2", "Conv2", "Conv3"]
        for node in nodes:
            self.assertTrue(node in nodes_name)

    def test_remove_nodes(self):
        nodes_to_remove = []
        for node in self.model.nodes():
            if node.name == "Conv3" or node.name == "Add":
                nodes_to_remove.append(node)
        self.model.remove_nodes(nodes_to_remove)
        self.assertEqual(len(self.model.nodes()), 4)
        nodes_name = [node.name for node in self.model.nodes()]
        nodes = ["Relu1", "Conv1", "Relu2", "Conv2"]
        for node in nodes:
            self.assertTrue(node in nodes_name)

    def test_add_node(self):
        node_to_add = onnx.helper.make_node('Relu', ['output'], ['output1'], keepdims=0)
        self.model.add_node(node_to_add)
        last_node = self.model.nodes()[-1]
        self.assertEqual(last_node.op_type, 'Relu')

    def test_add_nodes(self):
        nodes_to_add = []
        for i in range(2):
            node_to_add = onnx.helper.make_node('Relu', ["add_node{}_input".format(str(i))], ["add_node{}_output".format(str(i))], keepdims=0)
            nodes_to_add.append(node_to_add)
        self.model.add_nodes(nodes_to_add)
        self.assertEqual(self.model.nodes()[-1].input, ['add_node1_input'])
        self.assertEqual(self.model.nodes()[-2].input, ['add_node0_input'])
        self.assertEqual(self.model.nodes()[-1].output, ['add_node1_output'])
        self.assertEqual(self.model.nodes()[-2].output, ['add_node0_output'])

    def test_get_initializer(self):
        inits = ['X1_weight', 'X1_bias', 'X3_weight', 'X3_bias', 'X5_weight', 'X5_bias']
        for init in inits:
            self.assertIsNotNone(self.model.get_initializer(init))

    def test_remove_initializer(self):
        for init in self.model.initializer():
            if init.name == "X1_weight":
                self.model.remove_initializer(init)
        self.assertEqual(len(self.model.initializer()), 5)
        inits_name = [init.name for init in self.model.initializer()]
        inits = ['X1_bias', 'X3_weight', 'X3_bias', 'X5_weight', 'X5_bias']
        for init in inits:
            self.assertTrue(init in inits_name)

    def test_remove_initializers(self):
        init_to_remove = []
        for init in self.model.initializer():
            if "bias" in init.name:
                init_to_remove.append(init)
        self.model.remove_initializers(init_to_remove)
        self.assertEqual(len(self.model.initializer()), 3)
        inits_name = [init.name for init in self.model.initializer()]
        inits = ['X1_weight', 'X3_weight', 'X5_weight']
        for init in inits:
            self.assertTrue(init in inits_name)

    def test_input_name_to_nodes(self):
        self.assertEqual(len(self.model.input_name_to_nodes), 12)
        ipts_name = [name for name in self.model.input_name_to_nodes]
        ipts = ['input0', 'X1',  'X2', 'X3', 'X3_weight', 'X3_bias','X5_weight', 'X5_bias', 'X4', 'X5']
        for ipt in ipts:
            self.assertTrue(ipt in ipts_name)

    def test_output_name_to_node(self):
        self.assertEqual(len(self.model.output_name_to_node), 6)
        opts_name = [name for name in self.model.output_name_to_node]
        opts = ['X1', 'X2', 'X3', 'X4', 'X5', 'output']
        for opt in opts:
            self.assertTrue(opt in opts_name)
    
    def test_get_children(self):
        for node in self.model.nodes():
            if node.name == "Relu1":
                children = self.model.get_children(node)
        self.assertEqual(len(children), 2)
        children_name = [child.name for child in children]
        names = ["Conv1", "Conv3"]
        for name in names:
            self.assertTrue(name in children_name)

    def test_get_parents(self):
        for node in self.model.nodes():
            if node.op_type == "Add":
                parents = self.model.get_parents(node)
        self.assertEqual(len(parents), 2)
        parents_name = [parent.name for parent in parents]
        names = ["Conv2", "Conv3"]
        for name in names:
            self.assertTrue(name in parents_name)

    def test_get_parent(self):
        for node in self.model.nodes():
            if node.op_type == "Add":
                node_to_get_parent = node
        parent = self.model.get_parent(node, 0)
        self.assertEqual(parent.name, "Conv2")
        parent = self.model.get_parent(node, 1)
        self.assertEqual(parent.name, "Conv3")
        parent = self.model.get_parent(node, 2)
        self.assertIsNone(parent)

    def test_find_nodes_by_initializer(self):
        for init in self.model.initializer():
            if init.name == "X1_weight":
                initializer = init
        nodes = self.model.find_nodes_by_initializer(self.model.graph(), initializer)
        self.assertEqual(len(nodes), 1)
        self.assertEqual(nodes[0].name, "Conv1")

    def test_get_scale_zero(self):
        input_scale, input_zero = self.q_model.get_scale_zero('B_quantized')
        weight_scale, weight_zero = self.q_model.get_scale_zero('C_quantized') 
        bias_scale, bias_zero = self.q_model.get_scale_zero('E')

    def test_save(self):
        self.model.save_model_to_file('./test_model_6.onnx', use_external_data_format=True)

    
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Download config test."""

import unittest
from unittest.mock import MagicMock, patch

from neural_compressor.ux.components.model_zoo.download_config import download_config


@patch("sys.argv", ["neural_compressor_bench.py", "-p5000"])
class TestDownloadConfig(unittest.TestCase):
    """DownloadConfig tests."""

    @patch("neural_compressor.ux.components.model_zoo.download_config.Downloader")
    def test_download_config(self, downloader_mock: MagicMock) -> None:
        """Test download_config."""
        data = {
            "id": "some request id",
        }

        download_config(data)

        downloader_mock.assert_called_once_with(data)
        downloader_mock.return_value.download_config.assert_called_once()


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration to yaml."""

import logging
import os
from copy import deepcopy
from shutil import copy
from typing import Any, Dict, List, Optional, Tuple, Union

from neural_compressor.ux.components.configuration_wizard.configuration_parser import (
    ConfigurationParser,
)
from neural_compressor.ux.components.model.repository import ModelRepository
from neural_compressor.ux.components.optimization import Optimizations
from neural_compressor.ux.utils.exceptions import NotFoundException
from neural_compressor.ux.utils.hw_info import HWInfo
from neural_compressor.ux.utils.templates.workdir import Workdir
from neural_compressor.ux.utils.utils import replace_with_values
from neural_compressor.ux.utils.workload.config import Config
from neural_compressor.ux.utils.workload.dataloader import Dataloader, Dataset, Transform
from neural_compressor.ux.utils.workload.workload import ExecutionMode, Workload

logging.basicConfig(level=logging.INFO)


def set_dataloader_to_dummy(dataloader: Dataloader, shape: list) -> None:
    """Change dataloader configuration to dummy_v2."""
    dataloader.transform.clear()
    dataloader.dataset = Dataset(
        "dummy_v2",
        {
            "input_shape": deepcopy(shape),
            "label_shape": [1],
        },
    )


def change_performance_dataloader_to_dummy_if_possible(
    model_domain: str,
    config: Config,
) -> None:
    """Change config.evaluation.performance.dataloader.dataset to Dummy_v2."""
    model = ModelRepository().get_model(config.model_path)
    detected_shape = model.input_shape
    if (
        detected_shape.trusted
        and detected_shape.shape
        and config.evaluation
        and config.evaluation.performance
        and config.evaluation.performance.dataloader
    ):
        shape = ConfigurationParser.parse_value(detected_shape.shape, [int])
        set_dataloader_to_dummy(config.evaluation.performance.dataloader, shape)
        return

    if model_domain not in ["image_recognition", "object_detection"]:
        return

    if not (
        config.evaluation
        and config.evaluation.performance
        and config.evaluation.performance.dataloader
        and config.evaluation.performance.dataloader.dataset
    ):
        return

    if not (
        config.quantization
        and config.quantization.calibration
        and config.quantization.calibration.dataloader
        and config.quantization.calibration.dataloader.transform
    ):
        return

    try:
        transforms = [
            value for _, value in config.quantization.calibration.dataloader.transform.items()
        ]
        shape = get_shape_from_transforms(transforms, model.shape_elements_order)
        set_dataloader_to_dummy(config.evaluation.performance.dataloader, shape)
    except (NotFoundException, ValueError):
        pass


def get_shape_from_transforms(
    transforms: List[Transform],
    shape_elements_order: List[str],
) -> list:
    """Detect dataset sizes based on configured transforms."""
    shapes = {
        "channels": 3,
        "height": None,
        "width": None,
    }

    for transform in transforms:
        name = transform.name
        parameters = transform.parameters
        if name in [
            "Resize",
            "CenterCrop",
            "RandomResizedCrop",
            "RandomCrop",
            "CropResize",
        ]:
            shapes["height"], shapes["width"] = get_height_width_from_size(parameters.get("size"))
        elif "Transpose" == name:
            axes = parameters.get("perm")
            if not axes:
                raise ValueError("Unknown value of 'perm' argument in Transpose")
            shape_elements_order = [shape_elements_order[idx] for idx in axes]
        elif "CropToBoundingBox" == name:
            shapes["height"] = parameters.get("target_height")
            shapes["width"] = parameters.get("target_width")
        elif name in [
            "ResizeCropImagenet",
            "BilinearImagenet",
        ]:
            shapes["height"] = parameters.get("height")
            shapes["width"] = parameters.get("width")

    if not shapes["height"] or not shapes["width"]:
        raise NotFoundException("Unable to detect shape for Dummy dataset")

    return [shapes.get(dimension) for dimension in shape_elements_order]


def get_height_width_from_size(size: Any) -> Tuple[Optional[int], Optional[int]]:
    """Detect dataset sizes based on size param common in some Transforms."""
    if isinstance(size, int):
        return size, size
    elif isinstance(size, list):
        if len(size) == 1:
            return size[0], size[0]
        elif len(size) == 2:
            return size[0], size[1]
    return None, None


def change_evaluation_accuracy_configs_to_machine_specs(
    config: Config,
) -> None:
    """Change config.evaluation.accuracy.configs to match server."""
    if config.evaluation and config.evaluation.accuracy and config.evaluation.accuracy.configs:
        hwinfo = HWInfo()
        config.evaluation.accuracy.configs.cores_per_instance = hwinfo.cores_per_socket
        config.evaluation.accuracy.configs.num_of_instance = 1


def change_config_to_performance_mode(
    config: Config,
    data: dict,
) -> None:
    """Change config values to performance mode."""
    shape_definition = data.get("shape", "")
    shape = ConfigurationParser.parse_value(shape_definition, [int])
    if (
        config.quantization
        and config.quantization.calibration
        and config.quantization.calibration.dataloader
    ):
        set_dataloader_to_dummy(config.quantization.calibration.dataloader, shape)

    if (
        config.evaluation
        and config.evaluation.performance
        and config.evaluation.performance.dataloader
    ):
        set_dataloader_to_dummy(config.evaluation.performance.dataloader, shape)

    if config.evaluation:
        config.evaluation.accuracy = None


def save_workload(
    data: Dict[str, Any],
) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
    """Get configuration."""
    model_domain = data.get("domain", "")

    parser = ConfigurationParser()
    parsed_data = parser.parse(data)

    workload = Workload(parsed_data)
    workload.dump()

    workdir = Workdir(
        request_id=workload.id,
        project_name=workload.project_name,
        model_path=workload.model_path,
        input_precision=workload.input_precision,
        output_precision=workload.output_precision,
        mode=workload.mode,
        created_at=workload.created_at,
    )

    update_config(workload, parsed_data, workdir)

    if ExecutionMode.ADVANCED == workload.execution_mode:
        change_performance_dataloader_to_dummy_if_possible(model_domain, workload.config)
        change_evaluation_accuracy_configs_to_machine_specs(workload.config)
    else:
        change_config_to_performance_mode(workload.config, parsed_data)

    if Optimizations.GRAPH == workload.mode:
        workload.config.quantization = None

    workload.config.dump(os.path.join(workdir.workload_path, workload.config_name))
    return workload.serialize()


def check_custom_dataloader(data: dict) -> bool:
    """Check if custom dataloader is used in calibration or evaluation."""
    return any(
        [
            data.get("quantization", {}).get("dataloader", {}).get("name") == "custom",
            data.get("evaluation", {}).get("dataloader", {}).get("name") == "custom",
        ],
    )


def update_config(workload: Workload, data: dict, workdir: Workdir) -> None:
    """Update config data from request."""
    config = workload.config
    is_custom_dataloader = check_custom_dataloader(data)
    is_custom_metric = data.get("evaluation", {}).get("metric") == "custom"

    map_key_to_action: Dict[str, Any] = {
        "transform": config.set_transform,
        "model_path": config.set_model_path,
        "inputs": config.set_inputs,
        "outputs": config.set_outputs,
    }

    for key, action in map_key_to_action.items():
        if data.get(key, None) is not None:
            action(data[key])
    if data.get("evaluation", None):
        update_evaluation_data(config, data["evaluation"])
    if data.get("quantization", None):
        update_quantization_data(config, data["quantization"])
    if is_custom_metric and is_custom_dataloader:
        generate_template(workload, workdir, "dataloader_and_metric")
        config.remove_dataloader()
        config.remove_accuracy_metric()
    elif is_custom_dataloader:
        generate_template(workload, workdir, "dataloader")
        config.remove_dataloader()
    elif is_custom_metric:
        generate_template(workload, workdir, "metric")
        config.remove_accuracy_metric()


def generate_template(workload: Workload, workdir: Workdir, type: str) -> None:
    """Generate code templates."""
    correct_paths = {
        "config_path": workload.config_path,
        "model_path": workload.model_path,
        "model_output_path": workload.model_output_path,
    }

    generated_template_path = os.path.join(workload.workload_path, "code_template.py")
    path_to_templates = os.path.join(
        os.path.dirname(__file__),
        "..",
        "..",
        "utils",
        "templates",
        f"{type}_template.txt",
    )
    copy(path_to_templates, generated_template_path)
    replace_with_values(correct_paths, generated_template_path)
    workdir.set_code_template_path(generated_template_path)


def update_evaluation_data(config: Config, evaluation_data: Dict[str, Any]) -> None:
    """Update config with evaluation data."""
    config.set_accuracy_metric(evaluation_data)

    map_key_to_action: Dict[str, Any] = {
        "warmup": config.set_performance_warmup,
        "iterations": config.set_performance_iterations,
        "dataloader": config.set_evaluation_dataloader,
        "dataset_path": config.set_evaluation_dataset_path,
        "cores_per_instance": config.set_performance_cores_per_instance,
        "num_of_instance": config.set_performance_num_of_instance,
        "batch_size": config.set_accuracy_and_performance_batch_sizes,
    }

    for key, action in map_key_to_action.items():
        if evaluation_data.get(key, None):
            action(evaluation_data[key])


def update_quantization_data(config: Config, quantization_data: Dict[str, Any]) -> None:
    """Update config with quantization data."""
    map_key_to_action: Dict[str, Any] = {
        "accuracy_goal": config.set_accuracy_goal,
        "timeout": config.tuning.set_timeout,
        "max_trials": config.tuning.set_max_trials,
        "random_seed": config.tuning.set_random_seed,
        "approach": config.set_quantization_approach,
        "sampling_size": config.set_quantization_sampling_size,
        "dataloader": config.set_quantization_dataloader,
        "dataset_path": config.set_quantization_dataset_path,
        "batch_size": config.set_quantization_batch_size,
    }

    for key, action in map_key_to_action.items():
        if quantization_data.get(key, None):
            action(quantization_data[key])

    if quantization_data.get("strategy"):
        config.tuning.strategy.name = str(quantization_data["strategy"])
    if quantization_data.get("objective"):
        config.tuning.objective = str(quantization_data["objective"])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from collections import OrderedDict
from neural_compressor.utils import logger
import numpy as np
import yaml
import os


class Graph(object):
    def __init__(self):
        self._nodes = []
        self._node_id = {}
        self._engine = None

    @property
    def nodes(self):
        return self._nodes

    @nodes.setter
    def nodes(self, new_nodes):
        self._nodes = new_nodes

    def insert_nodes(self, index, nodes):
        idx = index
        for node in nodes:
            node = self.modify_node_connections(node, mode='insert')
            self._nodes.insert(idx, node)
            self._node_id[node.name] = idx
            for i in range(idx + 1, len(self._nodes)):
                self._node_id[self._nodes[i].name] += 1
            idx += 1
        self._engine = None

    def remove_nodes(self, node_names):
        for node_name in node_names:
            if node_name not in self._node_id.keys():
                continue
            node = self.get_node_by_name(node_name)
            _ = self.modify_node_connections(node, mode='remove')
            index = self.get_node_id(node_name)
            for i in range(index + 1, len(self._nodes)):
                self._node_id[self._nodes[i].name] -= 1
            self._nodes.pop(index)
            self._node_id.pop(node_name)
        self._engine = None

    def get_node_id(self, node_name):
        try:
            index = self._node_id[node_name]
            return index
        except BaseException:
            raise ValueError(
                'There is no node named {}, please check the input name.'.format(node_name))

    def get_node_by_name(self, node_name):
        index = self.get_node_id(node_name)
        if index is not None:
            return self._nodes[index]
        else:
            return None

    def rename_node(self, old_name, new_name):
        index = self.get_node_id(old_name)
        for i in range(len(self._nodes[index].input_tensors)):
            self._nodes[index].input_tensors[i].dest_op = [new_name]
            for pre_node_name in self._nodes[index].input_tensors[i].source_op:
                tensor_idx = self.get_tensor_idx(pre_node_name, 
                                            self._nodes[index].input_tensors[i].name)
                pre_node_idx = self._node_id[pre_node_name]
                self._nodes[pre_node_idx].output_tensors[tensor_idx].dest_op.remove(old_name)
                self._nodes[pre_node_idx].output_tensors[tensor_idx].dest_op.append(new_name)
        for i in range(len(self._nodes[index].output_tensors)):
            self._nodes[index].output_tensors[i].source_op = [new_name]
            for next_node_name in self._nodes[index].output_tensors[i].dest_op:
                tensor_idx = self.get_tensor_idx(next_node_name, 
                                self._nodes[index].output_tensors[i].name, from_output=False)
                next_node_idx = self._node_id[next_node_name]
                self._nodes[next_node_idx].input_tensors[tensor_idx].source_op = [new_name]

        self._nodes[index].name = new_name
        self._node_id.pop(old_name)
        self._node_id[new_name] = index
        self._engine = None

    def change_node_input_tensors(self, node_name, index, tensor=None, mode='modify'):
        assert mode in ['insert', 'remove', 'modify'], 'Wrong mode'
        node = self.get_node_by_name(node_name)
        index = index if index != -1 else len(node.input_tensors) - 1
        node_index = self.get_node_id(node_name)
        source_node_idx = None
        tensor_idx = None
        if mode == 'remove':
            tensor = node.input_tensors[index]
        assert tensor is not None
        tensor.dest_op = [node_name]
        if tensor.source_op != []:
            source_node_idx = self.get_node_id(tensor.source_op[0])
            tensor_idx = self.get_tensor_idx(tensor.source_op[0], tensor.name, from_output=True)

        if mode == 'insert':
            if source_node_idx is not None:
                if node_name not in \
                self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op:
                    self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op.append(
                        node_name)
            self._nodes[node_index].input_tensors.insert(index, tensor)
        elif mode == 'remove':
            if source_node_idx is not None:
                self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op.remove(node_name)
            self._nodes[node_index].input_tensors.pop(index)

        else:
            self.change_node_input_tensors(node_name, index, mode='remove')
            self.change_node_input_tensors(node_name, index, tensor=tensor, mode='insert')
        self._engine = None

    def change_node_output_tensors(self, node_name, index, tensor=None, mode='modify'):
        assert mode in ['insert', 'remove', 'modify'], 'Wrong mode'
        node = self.get_node_by_name(node_name)
        index = index if index != -1 else len(node.output_tensors) - 1
        node_index = self.get_node_id(node_name)
        if mode == 'remove':
            tensor = node.output_tensors[index]
        assert tensor is not None
        tensor.source_op = [node_name]

        if mode == 'insert':
            self._nodes[node_index].output_tensors.insert(index, tensor)
        elif mode == 'remove':
            self._nodes[node_index].output_tensors.pop(index)
        else:
            self._nodes[node_index].output_tensors[index] = tensor
        self._engine = None

    def get_pre_node_names(self, node_name):
        pre_node_names = []
        node = self.get_node_by_name(node_name)
        for input_tensor in node.input_tensors:
            if input_tensor.source_op != []:
                pre_node_names.extend(input_tensor.source_op)

        return pre_node_names

    def get_next_node_names(self, node_name):
        next_node_names = []
        node = self.get_node_by_name(node_name)
        for output_tensor in node.output_tensors:
            if output_tensor.dest_op != []:
                next_node_names.extend(output_tensor.dest_op)

        return next_node_names

    def get_tensor_idx(self, node_name, tensor_name, from_output=True):
        target_node = self.get_node_by_name(node_name)
        tensor_idx = -1
        if from_output:
            target_tensors = target_node.output_tensors
        else:
            target_tensors = target_node.input_tensors
        for j in range(len(target_tensors)):
            if target_tensors[j].name == tensor_name:
                tensor_idx = j
                break
            else:
                continue
        # assert tensor_idx != -1, 'Graph does not has tensor {}, '\
        #'please check it.'.format(tensor_name)

        return tensor_idx

    def modify_node_connections(self, node, mode='insert'):
        assert mode in ['insert', 'remove'], 'Wrong mode {}'.format(mode)
        # modify the input_tensors' source_op
        for i in range(len(node.input_tensors)):
            node.input_tensors[i].dest_op = [node.name]
            t = node.input_tensors[i]
            if t.source_op != [] and t.source_op[0] in self._node_id.keys():
                source_node_idx = self.get_node_id(t.source_op[0])
                source_node = self._nodes[source_node_idx]
                tensor_idx = self.get_tensor_idx(source_node.name, t.name)
                if mode == 'insert':
                    if node.name not in \
                    self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op:
                        self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op.append(
                            node.name)
                if mode == 'remove':
                    self._nodes[source_node_idx].output_tensors[tensor_idx].dest_op.remove(
                        node.name)
            # skip the const tensor and the node has been removed
            else:
                continue

        # modify the output_tensors' dest_op
        if mode == 'insert':
            for i in range(len(node.output_tensors)):
                node.output_tensors[i].source_op = [node.name]
                t = node.output_tensors[i]
                for dest_op_name in node.output_tensors[i].dest_op:
                    if dest_op_name in self._node_id.keys():
                        dest_node_idx = self.get_node_id(dest_op_name)
                        tensor_idx = self.get_tensor_idx(dest_op_name, t.name, from_output=False)
                        if tensor_idx != -1:
                            self._nodes[dest_node_idx].input_tensors[tensor_idx].source_op = [
                                node.name]
        self._engine = None

        return node

    # get the weight_bytes to bin file
    @property
    def weight_data(self):
        consts_info = OrderedDict()
        weight_bytes = bytearray()
        non_consts_len = 0
        for t in self._nodes[0].output_tensors:
            assert self._nodes[0].op_type=='Input', 'The graph must have input data'
            if t.source_op==[] and isinstance(t.data, np.ndarray):
                break
            else:
                non_consts_len += 1
        self._nodes[0].output_tensors = self._nodes[0].output_tensors[:non_consts_len]
        for i in range(len(self._nodes)):
            for j in range(len(self._nodes[i].input_tensors)):
                t = self._nodes[i].input_tensors[j]
                if t.source_op==[] and isinstance(t.data, np.ndarray):
                    data = t.data
                    start = len(weight_bytes)
                    data_bytes = data.tobytes()
                    weight_bytes.extend(data_bytes)
                    offset = len(data_bytes)
                    self._nodes[i].input_tensors[j].location = [start, offset]
                    self._nodes[0].output_tensors.append(self._nodes[i].input_tensors[j])
        weight_bytes = bytes(weight_bytes)
        return weight_bytes

    # get the network config dict to yaml file
    @property
    def net_config(self):
        net_info = OrderedDict()
        net_info['model'] = OrderedDict()
        net_info['model']['name'] = 'model'
        net_info['model']['operator'] = OrderedDict()
        for node in self._nodes:
            net_info['model']['operator'][node.name] = node.config

        return net_info

    def dump_tensor(self, tensor_list=[]):
        weight_data = self.weight_data
        net_info = self.net_config
        if tensor_list == []:
            for node in net_info['model']['operator']:
                if node != 'input_data' and 'output' in net_info['model']['operator'][node].keys():
                    for tensor in net_info['model']['operator'][node]['output']:
                        net_info['model']['operator']['output_data']['input'][tensor] = {}
        else:
            for tensor in tensor_list:
                for node in net_info['model']['operator']:
                    operator = net_info['model']['operator']
                    if 'output' not in operator[node].keys():
                        continue
                    for tensor_name in operator[node]['output']:
                        search = re.search(tensor, tensor_name, re.I)
                        if search is not None:
                            net_info['model']['operator']['output_data']['input'][tensor_name] = {}

        return net_info

    # pybind engine executor
    def engine_init(self, net_info={}, weight_data=b""):
        import engine_py as dp
        if not weight_data:
            weight_data = self.weight_data
        if not net_info:
            net_info = self.net_config
        op_configs = []
        tensor_output = []
        tensor_input = []
        attr_map_list = []
        for node in net_info['model']['operator']:
            tensor_input.append([])
            tensor_output.append([])
            opeartor = net_info['model']['operator'][node]
            if 'input' in opeartor.keys():
                for input_name in opeartor['input']:
                    input_tensor = dp.tensor_config(input_name, [], "fp32", [], [])
                    tensor_input[-1].append(input_tensor)

            if 'output' in opeartor.keys():
                for (output_name, attrs) in opeartor['output'].items():
                    tensor_location = []
                    if 'location' in attrs.keys():
                        tensor_location = attrs['location']
                    tensor_strides = []
                    if "strides" in attrs.keys():
                        tensor_strides = attrs["strides"]
                    tensor_shape = []
                    if "shape" in attrs.keys():
                        tensor_shape = attrs["shape"]  
                    tensor_dtype = 'fp32'
                    if "dtype" in attrs.keys():
                        tensor_dtype = attrs["dtype"]
                    output_tensor = dp.tensor_config(output_name, tensor_shape, tensor_dtype,
                                                     tensor_strides, tensor_location)
                    tensor_output[-1].append(output_tensor)

            if 'attr' in opeartor.keys():
                op_attr = opeartor['attr']
                attr_maps = {}
                for (k, v) in op_attr.items():
                    attr_maps[str(k)] = str(v)
                attr_map_item = dp.attrs_config(attr_maps)
                attr_map_list.append(attr_map_item)
            else:
                attr_map = dp.attrs_config({})
                attr_map_list.append(attr_map)
            op_type = net_info['model']['operator'][node]['type']
            op_config = dp.op_config(str(node), str(op_type), tensor_input[-1], tensor_output[-1],
                                     attr_map_list[-1])
            op_configs.append(op_config)

        model_config = dp.model_config(net_info['model']['name'], op_configs)
        output_list = []
        for node in net_info['model']['operator']['output_data']['input']:
            output_list.append(node)
        model = dp.Model(model_config, weight_data)
        self._engine = [model, output_list, op_configs, tensor_output, tensor_input, attr_map_list]

    def inference(self, input_data):
        if self._engine is None:
            self.engine_init()
        output = self._engine[0].forward(input_data)
        index = 0
        output_dict = OrderedDict()
        for node in self._engine[1]:
            output_dict[node] = output[index]
            index += 1

        return output_dict

    def graph_init(self, config, weight_data=None):
        '''
        example:
                from engine.converter.graph import Graph
                newgraph = Graph()
                newgraph.graph_init('./ir/conf.yaml', './ir/model.bin')
                out = newgraph.inference([input_0, input_1, input_2])
        '''
        from ..ops import Tensor
        import yaml
        from .. import graph_utils as util
        import copy
        self._nodes = []
        self._node_id = {}
        self._engine = None
        yamlPath = os.path.join(config)
        f = open(yamlPath, 'r', encoding='utf-8')
        cfg = f.read()
        d = yaml.load(cfg, Loader=yaml.FullLoader)
        bin_file = None
        if weight_data != None:
            bin_file = open(weight_data, 'rb')

        tensor_name_2_class = OrderedDict()
        for node in d['model']['operator']:
            op = None
            optype = d['model']['operator'][node]['type']
            if optype == 'Input':
                output_tensors = []
                for (tensor_name, attrs) in d['model']['operator'][node]['output'].items():
                    tensor_strides = None
                    if "strides" in attrs.keys():
                        tensor_strides = attrs["strides"]
                    tensor_shape = None
                    if "shape" in attrs.keys():
                        tensor_shape = attrs["shape"]
                    tensor_dtype = None
                    if "dtype" in attrs.keys():
                        tensor_dtype = attrs["dtype"]
                    tensor_location = None
                    tensor_data = None
                    if 'location' in attrs.keys():
                        tensor_location = attrs['location']
                        bin_file.seek(tensor_location[0], 0)
                        tensor_data = copy.deepcopy(bin_file.read(tensor_location[1]))
                        DTYPES_DICT = { "fp32": np.float32,
                                        "s8": np.int8,
                                        "s32": np.int32,
                                        "u8": np.uint8,
                                       }
                        tensor_data = np.frombuffer(tensor_data, dtype=DTYPES_DICT[tensor_dtype])
                    tensorclass = Tensor()
                    if tensor_location == None:
                        tensorclass = Tensor(tensor_name, ['input_data'], [], tensor_shape,
                                             tensor_data, tensor_dtype, tensor_location)
                    else:
                        tensorclass = Tensor(tensor_name, [], [], tensor_shape, tensor_data,
                                             tensor_dtype, tensor_location)
                    tensor_name_2_class[tensor_name] = tensorclass
                    output_tensors.append(tensorclass)
                op = util.construct_node(node, 'Input', [], copy.deepcopy(output_tensors))

            elif optype == 'Output':
                input_tensors = []
                for tensor_name in d['model']['operator'][node]['input']:
                    tensor = tensor_name_2_class[tensor_name]
                    input_tensors.append(tensor)
                op = util.construct_node(node, 'Output', copy.deepcopy(input_tensors))

            else:
                input_tensors = []
                output_tensors = []
                attr = {}
                for tensor_name in d['model']['operator'][node]['output']:
                    tensor = Tensor(tensor_name, [node])
                    tensor_name_2_class[tensor_name] = tensor
                    output_tensors.append(tensor)
                for tensor_name in d['model']['operator'][node]['input']:
                    tensor = tensor_name_2_class[tensor_name]
                    tensor.dest_op.append(node)
                    input_tensors.append(tensor)
                if 'attr' in d['model']['operator'][node].keys():
                    attr = d['model']['operator'][node]['attr']

                op = util.construct_node(node, optype, copy.deepcopy(input_tensors),
                                         copy.deepcopy(output_tensors), attr)
            self.insert_nodes(len(self.nodes), [op])

    def save(self, output_dir=None):
        logger.info("Start to emit the intermediate representation of model...")
        if output_dir is None:
            dir_name = os.getcwd()
            output_dir = os.path.join(dir_name, 'ir/')

        output_dir = os.path.abspath(output_dir)
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        # set the bin and yaml file name
        bin_file = os.path.join(output_dir, 'model.bin')
        yaml_file = os.path.join(output_dir, 'conf.yaml')

        # serialize_weight
        weight_data = self.weight_data
        with open(bin_file, 'wb') as f:
            f.write(weight_data)

        # serialize_network
        net_info = self.net_config
        with open(yaml_file, "w", encoding="utf-8") as f:
            # for write list, no use '-' to split the list, which is the default action in yaml
            def list_representer(dumper, data):
                return dumper.represent_sequence(u'tag:yaml.org,2002:seq', data, flow_style=True)
            # for write OrderedDict

            def dict_representer(dumper, data):
                return dumper.represent_mapping("tag:yaml.org,2002:map", data.items())
            yaml.add_representer(list, list_representer)
            yaml.add_representer(OrderedDict, dict_representer)
            yaml.dump(net_info, f, default_flow_style=False, sort_keys=False)

        logger.info("Emit done...")



import os
import unittest
import yaml
import numpy as np
import tensorflow as tf
import shutil
from tensorflow.python.framework import graph_util
from neural_compressor.adaptor.tensorflow import TensorFlowAdaptor
from neural_compressor.adaptor.tf_utils.util import disable_random

def build_fake_yaml():
    fake_yaml = '''
        model:
          name: fake_yaml
          framework: tensorflow
          inputs: input
          outputs: op_to_store
        device: cpu
        evaluation:
          accuracy:
            metric:
              topk: 1
        tuning:
            accuracy_criterion:
              relative: 0.0001
            workspace:
              path: saved
        '''
    y = yaml.load(fake_yaml, Loader=yaml.SafeLoader)
    with open('fake_yaml.yaml', "w", encoding="utf-8") as f:
        yaml.dump(y, f)
    f.close()
class TestSetTensor(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        build_fake_yaml()

    @classmethod
    def tearDownClass(self):
        os.remove('fake_yaml.yaml')
        shutil.rmtree('./saved', ignore_errors=True)
        
    @disable_random()
    def test_fp32bias(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(x, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                            initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(x_pad, conv_weights, strides=[1, 2, 2, 1], padding="VALID")

        conv_bias = tf.compat.v1.get_variable("bias", [16], dtype=tf.float32,
                                            initializer=tf.compat.v1.random_normal_initializer())

        conv_bias = tf.math.add(conv, conv_bias)
        relu6 = tf.nn.relu6(conv_bias, name='op_to_store')

        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            constant_graph = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization("./fake_yaml.yaml")
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = constant_graph
            q_model = quantizer()
            
            framework_specific_info = {'device': 'cpu', 'workspace_path': 'saved',\
                    'random_seed': 1978, 'inputs': ['input'], 'outputs': ['op_to_store'], \
                    'approach': 'post_training_static_quant'}
            adaptor = TensorFlowAdaptor(framework_specific_info)
            adaptor.set_tensor(q_model, {'bias': np.random.random(16)})

            from tensorflow.core.framework import attr_value_pb2
            from tensorflow.python.framework import dtypes
            for node in q_model.graph_def.node:
                if node.name == 'bias':
                    self.assertEqual(node.attr['dtype'], attr_value_pb2.AttrValue(
                    type=dtypes.float32.as_datatype_enum))

    @disable_random()
    def test_int32bias(self):
        x = tf.compat.v1.placeholder(tf.float32, [1, 56, 56, 16], name="input")
        paddings = tf.constant([[0, 0], [1, 1], [1, 1], [0, 0]])
        x_pad = tf.pad(x, paddings, "CONSTANT")
        conv_weights = tf.compat.v1.get_variable("weight", [3, 3, 16, 16],
                                            initializer=tf.compat.v1.random_normal_initializer())
        conv = tf.nn.conv2d(x_pad, conv_weights, strides=[1, 2, 2, 1], padding="VALID")

        conv_bias = tf.compat.v1.get_variable("bias", [16], dtype=tf.float32)

        conv_bias = tf.math.add(conv, conv_bias)
        relu6 = tf.nn.relu6(conv_bias, name='relu_0')

        conv_weights1 = tf.compat.v1.get_variable("weight1", [3, 3, 16, 16],
                                            initializer=tf.compat.v1.random_normal_initializer())
        conv1 = tf.nn.conv2d(relu6, conv_weights1, strides=[1, 2, 2, 1], padding="VALID")

        conv_bias1 = tf.compat.v1.get_variable("bias1", [16], dtype=tf.float32)

        conv_bias1 = tf.math.add(conv1, conv_bias1)
        relu6 = tf.nn.relu6(conv_bias1, name='relu_1')

        conv_weights2 = tf.compat.v1.get_variable("weight2", [3, 3, 16, 16],
                                            initializer=tf.compat.v1.random_normal_initializer())
        conv2 = tf.nn.conv2d(relu6, conv_weights2, strides=[1, 2, 2, 1], padding="VALID")

        conv_bias2 = tf.compat.v1.get_variable("bias2", [16], dtype=tf.float32)

        conv_bias2 = tf.math.add(conv2, conv_bias2)
        relu6 = tf.nn.relu6(conv_bias2, name='op_to_store')
        out_name = relu6.name.split(':')[0]
        with tf.compat.v1.Session() as sess:
            sess.run(tf.compat.v1.global_variables_initializer())
            constant_graph = graph_util.convert_variables_to_constants(
                sess=sess,
                input_graph_def=sess.graph_def,
                output_node_names=[out_name])

            for i in constant_graph.node:
                if i.op.find('Add') != -1:
                    i.op = 'Add'

            from neural_compressor.experimental import Quantization, common
            quantizer = Quantization("./fake_yaml.yaml")
            dataset = quantizer.dataset('dummy', shape=(100, 56, 56, 16), label=True)
            quantizer.calib_dataloader = common.DataLoader(dataset)
            quantizer.model = constant_graph
            q_model = quantizer()

            framework_specific_info = {'device': 'cpu', 'workspace_path': 'saved',\
                    'random_seed': 1978, 'inputs': ['input'], 'outputs': ['op_to_store'], \
                    'approach': 'post_training_static_quant'}
            adaptor = TensorFlowAdaptor(framework_specific_info)
            adaptor.set_tensor(q_model, {'bias1': np.random.randint(6,size=2, dtype='int32')})
            from tensorflow.core.framework import attr_value_pb2
            from tensorflow.python.framework import dtypes
            for node in q_model.graph_def.node:
                if node.name == 'bias2':
                    self.assertEqual(node.attr['dtype'], attr_value_pb2.AttrValue(
                    type=dtypes.qint32.as_datatype_enum))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import attribute_to_kwarg, ms_domain, \
                                                 QuantizedValueType
from neural_compressor.adaptor.ox_utils.util import QuantizedValue                                                 
class QGlobalAveragePool(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type == "GlobalAveragePool")

        # If input to this node is not quantized then keep this node.
        if node.input[0] not in self.quantizer.quantized_value_map:
            return super().quantize()
        quantized_input_value = self.quantizer.quantized_value_map[node.input[0]]

        # Create an entry for output quantized value.
        quantized_input_value = self.quantizer.quantized_value_map[node.input[0]]
        data_found, output_scale_name_from_parameter, output_zp_name_from_parameter, _, _ = \
            self.quantizer._get_quantization_params(node.output[0])
        # Just use input scale and zp if parameters for output is not specified.
        output_scale_name = output_scale_name_from_parameter if data_found else \
                                                             quantized_input_value.scale_name
        output_zp_name = output_zp_name_from_parameter if data_found else \
                                                       quantized_input_value.zp_name
        quantized_output_value = QuantizedValue(
            node.output[0], node.output[0] + "_quantized",
            output_scale_name, output_zp_name, quantized_input_value.qType)
        self.quantizer.quantized_value_map[node.output[0]] = quantized_output_value

        kwargs = {}
        for attribute in node.attribute:
            kwargs.update(attribute_to_kwarg(attribute))
        kwargs["domain"] = ms_domain
        kwargs["channels_last"] = 0
        qnode_name = node.name + "_quant" if node.name != "" else ""

        qnode = onnx.helper.make_node(
            "QLinear" + node.op_type,
            [quantized_input_value.q_name, quantized_input_value.scale_name, 
            quantized_input_value.zp_name, output_scale_name, output_zp_name],
            [quantized_output_value.q_name],
            qnode_name, **kwargs)
        self.quantizer.new_nodes += [qnode]



#
#  -*- coding: utf-8 -*-
#
import unittest
import os
import tensorflow as tf

from neural_compressor.adaptor.tf_utils.util import read_graph
from neural_compressor.adaptor.tf_utils.quantize_graph.quantize_graph_for_intel_cpu import QuantizeGraphForIntel
from neural_compressor.adaptor.tensorflow import TensorflowQuery
class TestTensorflowGpu(unittest.TestCase):
    mb_model_url = 'https://storage.googleapis.com/intel-optimized-tensorflow/models/v1_6/mobilenet_v1_1.0_224_frozen.pb'
    pb_path = '/tmp/.neural_compressor/mobilenet_fp32.pb'

    @classmethod
    def setUpClass(self):
        if not os.path.exists(self.pb_path):
            os.system("mkdir -p /tmp/.neural_compressor && wget {} -O {} ".format(self.mb_model_url, self.pb_path))
        self.op_wise_sequences = TensorflowQuery(local_config_file=os.path.join(
            os.path.dirname(__file__), "../neural_compressor/adaptor/tensorflow.yaml")).get_eightbit_patterns()

    def test_tensorflow_gpu_conversion(self):
        input_graph_def = read_graph(self.pb_path)
        output_node_names = ['MobilenetV1/Predictions/Reshape_1']
        op_wise_config = {
            'MobilenetV1/MobilenetV1/Conv2d_1_pointwise/Conv2D': (False, 'minmax', False, 7.0)}
        tf.compat.v1.disable_eager_execution()

        converter = QuantizeGraphForIntel(
            input_graph_def, output_node_names, op_wise_config, self.op_wise_sequences,  'gpu')
        converted_pb, _ = converter.do_transform()

        target_node_name = 'MobilenetV1/MobilenetV1/Conv2d_1_pointwise/Conv2D_eightbit_quantized_conv'

        node_details = {}
        for i in converted_pb.node:
            node_details[i.name] = i

        converted_flag = True if target_node_name in node_details else False

        self.assertEqual(converted_flag, True)

        target_node = node_details[target_node_name]
        weights_min_node = node_details[target_node.input[-2]]
        weights_max_node = node_details[target_node.input[-1]]

        self.assertEqual(weights_max_node.op, "HostConst")
        self.assertEqual(weights_min_node.op, "HostConst")

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Parsers for log files."""
import re
from abc import ABC
from typing import Any, Dict, List, Union

from neural_compressor.ux.components.benchmark import Benchmarks
from neural_compressor.ux.utils.exceptions import InternalException
from neural_compressor.ux.utils.logger import log
from neural_compressor.ux.utils.templates.metric import Metric


class Parser(ABC):
    """Parser abstract class."""

    def __init__(self, logs: list) -> None:
        """Initialize parser."""
        self._logs = logs
        self.metric = Metric()

    def process(self) -> Dict[str, Any]:
        """Process log files."""
        raise NotImplementedError

    @property
    def patterns(self) -> dict:
        """Set patterns to get metrics from lines."""
        raise NotImplementedError


class OptimizationParser(Parser):
    """Parser class is responsible for parsing optimization log files."""

    def process(self) -> Dict[str, Any]:
        """Process files."""
        for log_file in self._logs:
            log.debug(f"Read from {log_file}")

            with open(log_file) as f:
                for line in f:
                    for key in self.patterns:
                        prog = re.compile(self.patterns[key])
                        match = prog.search(line)
                        if match and match.groupdict().get(key):
                            requested_value = str(match.groupdict().get(key))
                            self.metric.insert_data(key, requested_value)
        parsed_data: Dict[str, Any] = self.metric.serialize()  # type: ignore
        return parsed_data

    @property
    def patterns(self) -> dict:
        """Set patterns to get metrics from lines."""
        return {
            "acc_input_model": r".*FP32 baseline is:\s+\[("
            r"(accuracy:\s+(?P<acc_input_model>(\d+(\.\d+)?)))?"
            r"(duration\s+\(seconds\):\s+(?P<duration>(\d+(\.\d+)?)))?"
            r"(memory footprint\s+\(MB\):\s+(?P<mem_footprint>(\d+(\.\d+)?)))?(,\s+)?"
            r")*\]",
            "acc_optimized_model": r".*Best tune result is:\s+\[("
            r"(accuracy:\s+(?P<acc_optimized_model>(\d+(\.\d+)?)))?"
            r"(duration\s+\(seconds\):\s+(?P<duration>(\d+(\.\d+)?)))?"
            r"(memory footprint\s+\(MB\):\s+(?P<mem_footprint>(\d+(\.\d+)?)))?(,\s+)?"
            r")*\]",
            "path_optimized_model": r".*Save quantized model to (?P<path_optimized_model>.*)\.",
        }


class PerformanceParser(Parser):
    """Parser class is responsible for parsing performance benchmark log files."""

    def process(self) -> Dict[str, Any]:
        """Process files."""
        partial: Dict[str, List] = {}
        for log_file in self._logs:
            log.debug(f"Read from {log_file}")

            with open(log_file) as f:
                for line in f:
                    for key in self.patterns:
                        prog = re.compile(self.patterns[key])
                        match = prog.search(line)
                        if not match:
                            continue
                        metric_name = f"perf_{key}_input_model"
                        self.metric.insert_data(metric_name, match.group(1))
                        converted_value = getattr(self.metric, metric_name)
                        parse_result = {
                            key: converted_value,
                        }
                        partial = self.update_partial(partial, parse_result)

        return self.summarize_partial(partial)

    @staticmethod
    def update_partial(
        partial: Dict[str, List],
        parsed_result: Dict[str, Union[float, int]],
    ) -> Dict[str, List]:
        """Update partial entries."""
        for key, value in parsed_result.items():
            if key not in partial:
                partial[key] = []
            partial[key].append(value)
        return partial

    def summarize_partial(self, partial: dict) -> dict:
        """Calculate final values."""
        summary = {}
        for key, value in partial.items():
            summarized_value = self.summarize_value(key, value)
            for precision in ["input_model", "optimized_model"]:
                metric_name = f"perf_{key}_{precision}"

                summary[metric_name] = summarized_value
        return summary

    @staticmethod
    def summarize_value(key: str, value: list) -> Union[float, int]:
        """Calculate final value."""
        if key == "latency":
            return round(sum(value) / len(value), 4)
        if key == "throughput":
            return round(sum(value), 4)
        return value[0]

    @property
    def patterns(self) -> dict:
        """Set patterns to get metrics from lines."""
        return {
            "throughput": r"Throughput:\s+(\d+(\.\d+)?)",
            "latency": r"Latency:\s+(\d+(\.\d+)?)",
        }


class AccuracyParser(Parser):
    """Parser class is responsible for parsing accuracy benchmark log files."""

    def process(self) -> Dict[str, Any]:
        """Process accuracy logs."""
        for log_file in self._logs:
            log.debug(f"Read from {log_file}")

            with open(log_file) as f:
                for line in f:
                    for key in self.patterns:
                        prog = re.compile(self.patterns[key])
                        match = prog.search(line)
                        if match:
                            for precision in ["input_model", "optimized_model"]:
                                metric_name = f"acc_{precision}"
                                self.metric.insert_data(metric_name, match.group(1))

        parsed_data: Dict[str, Any] = self.metric.serialize()  # type: ignore
        return parsed_data

    @property
    def patterns(self) -> dict:
        """Set patterns to get metrics from lines."""
        return {
            Benchmarks.ACC: r"Accuracy is (\d+(\.\d+)?)",
        }


class BenchmarkParserFactory:
    """Benchmark parser factory."""

    @staticmethod
    def get_parser(benchmark_mode: str, logs: List[str]) -> Parser:
        """Get benchmark parser for specified mode."""
        parser_map = {
            Benchmarks.PERF: PerformanceParser,
            Benchmarks.ACC: AccuracyParser,
        }
        parser = parser_map.get(benchmark_mode, None)
        if parser is None:
            raise InternalException(f"Could not find optimization class for {benchmark_mode}")
        return parser(logs)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import copy
import logging
from pathlib import Path
from neural_compressor.model.base_model import BaseModel

logger = logging.getLogger()

class EngineModel(BaseModel):
    def __init__(self, model, **kwargs):
        self._graph = None
        self._model = model
        self.q_config = None

    def framework(self):
        return 'engine'
    
    def _load_graph(self, model):
        if os.path.isdir(model):
            from engine.converter.graph import Graph
            file_list = os.listdir(model)
            assert len(file_list) == 2
            for file_name in file_list:
                file_ext= os.path.splitext(file_name)
                front, ext = file_ext
                if ext == ".yaml":
                    yaml_path = os.path.join(model, file_name)
                elif ext == ".bin":
                    bin_path = os.path.join(model, file_name)
                else:
                    logger.error("Please Input yaml and bin for engine.")
            graph = Graph()
            graph.graph_init(yaml_path, bin_path)
            logger.warn("When input yaml and bin, it can not use func in converter.")
        else:
            from engine.converter.loaders.loader import Loader
            from engine.converter.extractors.extractor import Extractor
            from engine.converter.sub_graph.subgraph_matcher import SubGraphMatcher
            loader = Loader()
            extractor = Extractor()
            sub_graph_matcher = SubGraphMatcher()
            graph = loader(model)
            graph = extractor(graph)
            graph = sub_graph_matcher(graph)
        return graph

    @property
    def graph(self):
        if not self._graph:
            self._graph = self._load_graph(self.model)
        return self._graph

    @graph.setter
    def graph(self, graph):
        self._graph = graph

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, model):
        self._model = model

    def save(self, output_dir=None):
        self.graph.save(output_dir)

    @property
    def nodes(self):
        return self.graph.nodes

    @nodes.setter
    def nodes(self, new_nodes):
        self.graph.nodes = new_nodes

    def insert_nodes(self, index, nodes):
        self.graph.insert_nodes(index, nodes)

    def remove_nodes(self, node_names):
        self.graph.remove_nodes(node_names)
    
    def get_node_id(self, node_name):
        return self.graph.get_node_id(node_name)

    def get_node_by_name(self, node_name):
        return self.graph.get_node_by_name(node_name)

    def rename_node(self, old_name, new_name):
        self.graph.rename_node(old_name, new_name)

    def get_pre_node_names(self, node_name):
        return self.graph.get_pre_node_names(node_name)

    def get_next_node_names(self, node_name):
        return self.graph.get_next_node_names(node_name)

    def modify_node_connections(self, node, mode='insert'):
        return self.graph.modify_node_connections(node, mode)

    def dump_tensor(self, tensor_list = []):
        return self.graph.dump_tensor(tensor_list)

    def engine_init(self, net_info={}, weight_data=b""):
        return self.graph.engine_init(net_info={}, weight_data=b"")

    def inference(self, input_data):
        return self.graph.inference(input_data)

    def change_node_input_tensors(self, node_name, index, tensor=None, mode='modify'):
        return self.graph.change_node_input_tensors(node_name, index, tensor, mode)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import onnx
from .base_operator import QuantOperatorBase
from onnxruntime.quantization.quant_utils import find_by_name, get_mul_node, \
                                                 QuantizedValueType, attribute_to_kwarg
from onnx import onnx_pb as onnx_proto
from neural_compressor.adaptor.ox_utils.util import QuantizedValue

class ConvInteger(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert node.op_type in ["Conv", "FusedConv"]

        (quantized_input_names, zero_point_names, scale_names, nodes) = \
                                        self.quantizer.quantize_inputs(node, [0, 1])

        # quantize bias if exist
        quantized_bias_name = ""
        bias_present = False
        if len(node.input) == 3:
            quantized_bias_name = self.quantizer.quantize_bias(node, nodes)
            bias_present = True

        conv_integer_output = node.output[0] + "_output_quantized"
        conv_integer_name = node.name + "_quant" if node.name != "" else ""

        kwargs = {}
        for attribute in node.attribute:
            if attribute.name == 'activation' and attribute.s in [b'Relu', b'Clip']:
                continue
            if attribute.name == 'activation_params':
                continue
            kwargs.update(attribute_to_kwarg(attribute))
        conv_integer_node = onnx.helper.make_node("ConvInteger", 
                                                  quantized_input_names + zero_point_names,
                                                  [conv_integer_output], 
                                                  conv_integer_name, **kwargs)
        nodes.append(conv_integer_node)

        # Add bias add nodes
        if bias_present:
            conv_integer_output = self.quantizer.get_bias_add_nodes(nodes, node, 
                                                                    conv_integer_output,
                                                                    quantized_bias_name)

        # Add cast operation to cast convInteger output to float.
        cast_op_output = conv_integer_output + "_cast_output"
        cast_node = onnx.helper.make_node("Cast", [conv_integer_output], [cast_op_output],
                                          conv_integer_output + "_cast",
                                          to=onnx_proto.TensorProto.FLOAT)
        nodes.append(cast_node)

        # Add mul operation to multiply scales of two inputs.
        assert (len(scale_names) == 2)
        if conv_integer_name != "":
            scales_mul_op = conv_integer_name + "_scales_mul"
        else:
            scales_mul_op = scale_names[0] + "_" + scale_names[1] + "_mul"

        scales_mul_node = find_by_name(scales_mul_op, self.quantizer.new_nodes)
        if scales_mul_node is None:
            scales_mul_node = get_mul_node(scale_names, scales_mul_op + ":0", scales_mul_op)
            nodes.append(scales_mul_node)

        scales_mul_op_output = scales_mul_node.output[0]

        # Add mul operation to multiply mul_scales_op result with output of ConvInteger
        # and make the output of this node the same as output of original conv node.
        output_scale_mul_op = conv_integer_name + "_output_scale_mul" if \
                                                  conv_integer_name != "" else ""
        nodes.append(get_mul_node([cast_op_output, scales_mul_op_output], 
                                  node.output[0], output_scale_mul_op))

        self.quantizer.new_nodes += nodes


class QLinearConv(QuantOperatorBase):
    def __init__(self, onnx_quantizer, onnx_node):
        super().__init__(onnx_quantizer, onnx_node)

    def quantize(self):
        node = self.node
        assert (node.op_type in ["Conv", "FusedConv"])

        if self.quantizer.is_input_a_weight(node.input[1]) and self.per_channel:
            (quantized_input_names, zero_point_names, scale_names, nodes) = \
                                                self.quantizer.quantize_inputs(node, [0])
            quant_weight_tuple = self.quantizer.quantize_weight_per_channel(node.input[1], 
                                                self.weight_dtype, self.weight_scheme, 0)
            quantized_input_names.append(quant_weight_tuple[0])
            zero_point_names.append(quant_weight_tuple[1])
            scale_names.append(quant_weight_tuple[2])
        else:
            (quantized_input_names, zero_point_names, scale_names, nodes) = \
                                            self.quantizer.quantize_inputs(node, [0, 1])

        quantized_bias_name = ""
        bias_present = False
        if len(node.input) == 3:
            quantized_bias_name = self.quantizer.quantize_bias(node, nodes)
            bias_present = True
        data_found, output_scale_name, output_zp_name, _, _ = \
            self.quantizer._get_quantization_params(node.output[0])

        if not data_found:
            raise ValueError("Quantization parameters for output:\"{}\" of node:\"{}\" not \
                              specified".format(node.output[0], node.name))

        qlinear_conv_output = node.output[0] + "_quantized"
        qlinear_conv_name = node.name + "_quant" if node.name != "" else ""

        kwargs = {}
        for attribute in node.attribute:
            if attribute.name == 'activation' and attribute.s in [b'Relu', b'Clip']:
                continue
            if attribute.name == 'activation_params':
                continue
            kwargs.update(attribute_to_kwarg(attribute))
        qlinear_conv_inputs = []
        # Input 0
        qlinear_conv_inputs.append(quantized_input_names[0])
        qlinear_conv_inputs.append(scale_names[0])
        qlinear_conv_inputs.append(zero_point_names[0])
        # Input 1
        qlinear_conv_inputs.append(quantized_input_names[1])
        qlinear_conv_inputs.append(scale_names[1])
        qlinear_conv_inputs.append(zero_point_names[1])

        # Output
        qlinear_conv_inputs.append(output_scale_name)
        qlinear_conv_inputs.append(output_zp_name)

        if bias_present:
            qlinear_conv_inputs.append(quantized_bias_name)

        qlinear_conv_node = onnx.helper.make_node("QLinearConv", qlinear_conv_inputs, 
                                                  [qlinear_conv_output],
                                                  qlinear_conv_name, **kwargs)
        nodes.append(qlinear_conv_node)

        # Create an entry for this quantized value
        q_output = QuantizedValue(node.output[0], qlinear_conv_output, 
                                  output_scale_name, output_zp_name,
                                  QuantizedValueType.Input)
        self.quantizer.quantized_value_map[node.output[0]] = q_output

        self.quantizer.new_nodes += nodes



"""Tests for Distributed TensorFlow Dataloader."""
from neural_compressor import data
from neural_compressor.utils.create_obj_from_config import create_dataset, create_dataloader
from neural_compressor.data.dataloaders.dataloader import DataLoader
from neural_compressor.data import DATASETS, DATALOADERS, TRANSFORMS
import tensorflow as tf
import numpy as np
import collections
import json
import os
import unittest
import shutil

class TestDistributedTFDataDataloader(unittest.TestCase):
    def setUp(self):
        tf.compat.v1.enable_eager_execution()
        self.dataset = tf.data.Dataset.from_tensors((tf.ones([3, 224, 224]), tf.ones([1000]))).repeat(600)
        self.count = 0

    def check_tf_dataset_with_batch_raise(self, batch_size, last_batch, distributed):
        dataset_with_batch = tf.data.Dataset.from_tensors\
            ((tf.ones([3, 224, 224]), tf.ones([1000]))).repeat(600).batch(2)
        dataloader = DATALOADERS['tensorflow']\
            (dataset_with_batch, batch_size=batch_size, last_batch=last_batch, distributed=distributed)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def check_distributed_raise(self, batch_size, last_batch, distributed):
        dataloader = DATALOADERS['tensorflow']\
            (self.dataset, batch_size=batch_size, last_batch=last_batch, distributed=distributed)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_1(self):
        self.assertRaises(TypeError, self.check_tf_dataset_with_batch_raise, 32, 'rollover', True)
    
    def test_dis_tf_data_dataloader_2(self):
        self.assertRaises(TypeError, self.check_tf_dataset_with_batch_raise, 32, 'no_rollover', True)

    def test_dis_tf_data_dataloader_3(self):
        self.assertRaises(TypeError, self.check_tf_dataset_with_batch_raise, 1, 'rollover', True)
    
    def test_dis_tf_data_dataloader_4(self):
        self.assertRaises(TypeError, self.check_tf_dataset_with_batch_raise, 1, 'no_rollover', True)

    def test_dis_tf_data_dataloader_5(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'rollover', True)
    
    def test_dis_tf_data_dataloader_6(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'no_rollover', True)

    def test_dis_tf_data_dataloader_7(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'rollover', True)
    
    def test_dis_tf_data_dataloader_8(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'no_rollover', True)

    def test_dis_tf_data_dataloader_9(self):
        batch_size = 32
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_10(self):
        batch_size = 32
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='no_rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_11(self):
        batch_size = 1
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_12(self):
        batch_size = 1
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='no_rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_13(self):
        batch_size = 600
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_dis_tf_data_dataloader_14(self):
        batch_size = 600
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch='no_rollover')
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (3, 224, 224))
            self.assertEqual(x[-1].shape, (3, 224, 224))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

class TestDefaultDataLoaderSequentialSampler(unittest.TestCase):
    @classmethod
    def tearDownClass(cls):
        if os.path.exists('minist'):
            shutil.rmtree('minist')
    
    def setUp(self):
        self.count = 0

    def check_get_len_raise(self, batch_size, last_batch, distributed):
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': last_batch,
            'distributed': distributed
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        len_dataloader = len(dataloader)

    def check_distributed_raise(self, batch_size, last_batch, distributed):
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': last_batch,
            'distributed': distributed
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler1(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 32, 'rollover', True)

    def test_sequential_sampler2(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 32, 'no_rollover', True)

    def test_sequential_sampler3(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 1, 'rollover', True)

    def test_sequential_sampler4(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 1, 'no_rollover', True)

    def test_sequential_sampler5(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'rollover', True)

    def test_sequential_sampler6(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'no_rollover', True)

    def test_sequential_sampler7(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'rollover', True)

    def test_sequential_sampler8(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'no_rollover', True)

    def test_sequential_sampler9(self):
        batch_size = 3332
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler10(self):
        batch_size = 3332
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'no_rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler11(self):
        batch_size = 1
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler12(self):
        batch_size = 1
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'no_rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler13(self):
        batch_size = 10000
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

    def test_sequential_sampler14(self):
        batch_size = 10000
        dataloader_args = {
            'batch_size': batch_size,
            'dataset': {"MNIST": {'root': './minist', 'train':False, 'download':True}},
            'transform': {'Resize': {'size': 24}},
            'filter': None,
            'last_batch': 'no_rollover'
        }
        dataloader = create_dataloader('tensorflow', dataloader_args)
        for batch in dataloader:
            x, y = batch
            if self.count < len(dataloader)-1:
                self.assertEqual(len(x), batch_size) 
            else:
                self.assertTrue(len(x) == dataloader.dis_dataset_len % batch_size or len(x) == batch_size)
            self.assertEqual(x[0].shape, (24, 24))
            self.assertEqual(x[-1].shape, (24, 24))
            self.assertIsInstance(x, np.ndarray)
            self.count += 1

class TestDefaultDataLoaderIterableSampler(unittest.TestCase):
    class iter_dataset(object):
        def __iter__(self):
            sample_size = 250
            for i in range(1, sample_size+1):
                yield np.array([i])

    def setUp(self):
        self.rank = 0
        self.size = 1 
        self.count = 1
        self.dataset = self.iter_dataset()

    def check_get_len_raise(self, batch_size, last_batch, distributed):
        dataloader = DATALOADERS['tensorflow']\
            (self.dataset, batch_size=batch_size, last_batch=last_batch, distributed=distributed)
        len_dataloader = len(dataloader)

    def check_distributed_raise(self, batch_size, last_batch, distributed):
        dataloader = DATALOADERS['tensorflow']\
            (self.dataset, batch_size=batch_size, last_batch=last_batch, distributed=distributed)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1
    
    def test_iterable_sampler1(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 32, 'rollover', True)

    def test_iterable_sampler2(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 32, 'no_rollover', True)

    def test_iterable_sampler3(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 1, 'rollover', True)

    def test_iterable_sampler4(self):
        self.assertRaises(EnvironmentError, self.check_get_len_raise, 1, 'no_rollover', True)

    def test_iterable_sampler5(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'rollover', True)

    def test_iterable_sampler6(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 32, 'no_rollover', True)

    def test_iterable_sampler7(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'rollover', True)

    def test_iterable_sampler8(self):
        self.assertRaises(EnvironmentError, self.check_distributed_raise, 1, 'no_rollover', True)

    def test_iterable_sampler9(self):
        batch_size = 128
        last_batch = 'rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

    def test_iterable_sampler10(self):
        batch_size = 128
        last_batch = 'no_rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

    def test_iterable_sampler11(self):
        batch_size = 1
        last_batch = 'rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

    def test_iterable_sampler12(self):
        batch_size = 1
        last_batch = 'no_rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

    def test_iterable_sampler13(self):
        batch_size = 1000
        last_batch = 'rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

    def test_iterable_sampler14(self):
        batch_size = 1000
        last_batch = 'no_rollover'
        dataloader = DATALOADERS['tensorflow'](self.dataset, batch_size=batch_size, last_batch=last_batch)
        for batch in dataloader:
            if self.count < len(dataloader):
                self.assertEqual(len(batch), batch_size)
                self.assertEqual(self.count*batch_size*self.size-self.size+self.rank+1, batch[-1][0])
            else:
                self.assertTrue(len(batch) == dataloader.dis_dataset_len % batch_size or len(batch) == batch_size)
                self.assertEqual(((self.count-1)*batch_size+len(batch)-1)*self.size+self.rank+1, batch[-1][0])
                break
            self.count += 1

class TestTensorflowBertDataLoader(unittest.TestCase):
    label = [{
        "paragraphs0":[
            {'context': 
                'Super Bowl 50 was an American football game to determine the champion of the National Football League (NFL) for the 2015 season.',
            'qas': [{
                'answers': [
                    {'answer_start': 177, 'text': 'Denver Broncos'}, 
                    {'answer_start': 177, 'text': 'Denver Broncos'}, 
                    {'answer_start': 177, 'text': 'Denver Broncos'}], 
                'question': 'Which NFL team represented the AFC at Super Bowl 50?', 
                'id': '56be4db0acb8001400a502ec'}]
            }
        ]
    }]
    unique_id = 1000000000
    input_ids = [101, 2029, 5088, 2136, 3421, 1996, 10511, 2012, 3565, 4605, 2753, 1029, 102, 3565, 4605, 2753,\
        1007, 2005, 1996, 2325, 2161, 1012, 1996, 2137, 2374, 3034, 1006]
    input_mask = [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    segment_ids = [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]
    fake_json = json.dumps({'data': label, 'version': '1.1'})
    with open('dev.json', 'w') as f:
        f.write(fake_json)   
    
    @classmethod
    def tearDownClass(cls):
        os.remove('test.record')
        os.remove('dev.json')
    
    def check_not_implement(self, batch_size, distributed):
        with tf.io.TFRecordWriter('./test.record') as writer:
            features = collections.OrderedDict()
            features["unique_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list([self.unique_id])))
            features["input_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_ids)))
            features["input_mask"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_mask)))
            features["segment_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.segment_ids)))
            tf_example = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(tf_example.SerializeToString())
        eval_dataset = create_dataset(
            'tensorflow',
            {'bert':{'root':'test.record', 'label_file': './dev.json'}},
            None,
            None)
        dataloader = DATALOADERS['tensorflow']\
            (dataset=eval_dataset, batch_size=batch_size, distributed=distributed)

    def test_tf_bert_dataloader_1(self):
        self.assertRaises(NotImplementedError, self.check_not_implement, 32, True)
    
    def test_tf_bert_dataloader_2(self):
        batch_size = 128
        with tf.io.TFRecordWriter('./test.record') as writer:
            features = collections.OrderedDict()
            features["unique_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list([self.unique_id])))
            features["input_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_ids)))
            features["input_mask"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_mask)))
            features["segment_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.segment_ids)))
            tf_example = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(tf_example.SerializeToString())
        eval_dataset = create_dataset(
            'tensorflow',
            {'bert':{'root':'test.record', 'label_file': './dev.json'}},
            None,
            None)
        dataloader = DATALOADERS['tensorflow'](dataset=eval_dataset, batch_size=batch_size)
        for inputs, labels in dataloader:
            self.assertEqual(inputs[0], 'test.record')
            self.assertEqual(inputs[1], batch_size)
            self.assertEqual(len(labels), 1)

    def test_tf_bert_dataloader_3(self):
        batch_size = 1
        with tf.io.TFRecordWriter('./test.record') as writer:
            features = collections.OrderedDict()
            features["unique_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list([self.unique_id])))
            features["input_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_ids)))
            features["input_mask"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.input_mask)))
            features["segment_ids"] = tf.train.Feature(
                int64_list=tf.train.Int64List(value=list(self.segment_ids)))
            tf_example = tf.train.Example(features=tf.train.Features(feature=features))
            writer.write(tf_example.SerializeToString())
        eval_dataset = create_dataset(
            'tensorflow',
            {'bert':{'root':'test.record', 'label_file': './dev.json'}},
            None,
            None)
        dataloader = DATALOADERS['tensorflow'](dataset=eval_dataset, batch_size=batch_size)
        for inputs, labels in dataloader:
            self.assertEqual(inputs[0], 'test.record')
            self.assertEqual(inputs[1], batch_size)
            self.assertEqual(len(labels), 1)

if __name__ == '__main__':
    unittest.main()


# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration to yaml."""

import json
from typing import Any, Dict

from neural_compressor.ux.components.model.repository import ModelRepository
from neural_compressor.ux.utils.exceptions import ClientErrorException, NotFoundException
from neural_compressor.ux.utils.logger import log
from neural_compressor.ux.utils.utils import check_module
from neural_compressor.ux.web.communication import MessageQueue

mq = MessageQueue()


def get_boundary_nodes(data: Dict[str, Any]) -> None:
    """Get configuration."""
    request_id = str(data.get("id", ""))
    model_path = data.get("model_path", None)

    if not (request_id and model_path):
        message = "Missing model path or request id."
        mq.post_error(
            "boundary_nodes_finish",
            {"message": message, "code": 404, "id": request_id},
        )
        return

    try:
        mq.post_success(
            "boundary_nodes_start",
            {"message": "started", "id": request_id},
        )
        model_repository = ModelRepository()
        try:
            model = model_repository.get_model(model_path)
        except NotFoundException:
            supported_frameworks = model_repository.get_frameworks()
            raise ClientErrorException(
                f"Framework for specified model is not yet supported. "
                f"Supported frameworks are: {', '.join(supported_frameworks)}.",
            )

        framework = model.get_framework_name()
        try:
            check_module(framework)
        except ClientErrorException:
            raise ClientErrorException(
                f"Detected {framework} model. "
                f"Could not find installed {framework} module. "
                f"Please install {framework}.",
            )

        response_data = {
            "id": request_id,
            "framework": framework,
            "inputs": model.get_input_nodes(),
            "outputs": model.get_output_nodes(),
        }

        response_data.update(model.domain.serialize())  # type: ignore
        response_data.update(model.input_shape.serialize())  # type: ignore
    except ClientErrorException as err:
        mq.post_error(
            "boundary_nodes_finish",
            {"message": str(err), "code": 404, "id": request_id},
        )
        return
    log.debug(f"Parsed data is {json.dumps(response_data)}")
    mq.post_success("boundary_nodes_finish", response_data)



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Metric class."""

from typing import Optional

from neural_compressor.ux.utils.json_serializer import JsonSerializer

DIGITS = 4


class Metric(JsonSerializer):
    """Metric represents data which is sent from Optimization and Benchmark."""

    # TODO: Split into Accuracy, Performance if necessary for Benchmark

    def __init__(self) -> None:
        """Initialize configuration Dataset class."""
        super().__init__()
        self._acc_input_model: Optional[float] = None
        self._acc_optimized_model: Optional[float] = None
        self._perf_latency_input_model: Optional[float] = None
        self._perf_latency_optimized_model: Optional[float] = None
        self._perf_throughput_optimized_model: Optional[float] = None
        self._perf_throughput_input_model: Optional[float] = None
        self._optimization_time: Optional[float] = None
        self._size_input_model: Optional[float] = None
        self._size_optimized_model: Optional[float] = None
        self._path_optimized_model: Optional[str] = None

    @property
    def acc_input_model(self) -> Optional[float]:
        """Accuracy for float32."""
        return self._acc_input_model

    @acc_input_model.setter
    def acc_input_model(self, value: str) -> None:
        """Set input model accuracy from value."""
        float_value = float(value)
        if float_value > 1:
            float_value /= 100
        self._acc_input_model = round(float_value, DIGITS)

    @property
    def acc_optimized_model(self) -> Optional[float]:
        """Accuracy for optimized model."""
        return self._acc_optimized_model

    @acc_optimized_model.setter
    def acc_optimized_model(self, value: str) -> None:
        """Set optimized model accuracy from value."""
        float_value = float(value)
        if float_value > 1:
            float_value /= 100
        self._acc_optimized_model = round(float_value, DIGITS)

    @property
    def perf_latency_input_model(self) -> Optional[float]:
        """Latency for input model."""
        return self._perf_latency_input_model

    @perf_latency_input_model.setter
    def perf_latency_input_model(self, value: str) -> None:
        """Set latency for input model."""
        self._perf_latency_input_model = round(float(value), DIGITS)
        if not self.perf_throughput_input_model:
            self.perf_throughput_input_model = self.calculate_throughput(
                self._perf_latency_input_model,
            )

    @property
    def perf_latency_optimized_model(self) -> Optional[float]:
        """Latency for optimized model."""
        return self._perf_latency_optimized_model

    @perf_latency_optimized_model.setter
    def perf_latency_optimized_model(self, value: str) -> None:
        """Set latency for optimized model."""
        self._perf_latency_optimized_model = round(float(value), DIGITS)
        if not self.perf_throughput_optimized_model:
            self.perf_throughput_optimized_model = self.calculate_throughput(
                self._perf_latency_optimized_model,
            )

    @property
    def perf_throughput_optimized_model(self) -> Optional[float]:
        """Throughput for optimized model."""
        return self._perf_throughput_optimized_model

    @perf_throughput_optimized_model.setter
    def perf_throughput_optimized_model(self, value: str) -> None:
        """Set throughput from value for optimized model."""
        self._perf_throughput_optimized_model = round(float(value), DIGITS)

    @property
    def perf_throughput_input_model(self) -> Optional[float]:
        """Throughput for input model."""
        return self._perf_throughput_input_model

    @perf_throughput_input_model.setter
    def perf_throughput_input_model(self, value: str) -> None:
        """Set throughput from value for input model."""
        self._perf_throughput_input_model = round(float(value), DIGITS)

    def insert_data(self, attribute: str, value: str) -> None:
        """Set attribute value."""
        self.__setattr__(attribute, value)

    @staticmethod
    def calculate_throughput(value: float) -> float:
        """
        Calculate throughput based on latency.

        Right now 1000 represents number of images in dataset.
        TODO: change 1000 to the batch size when Benchmark is ready
        """
        return 1000 / value

    @property
    def size_input_model(self) -> Optional[float]:
        """Model size for float32."""
        return self._size_input_model

    @size_input_model.setter
    def size_input_model(self, value: str) -> None:
        """Set input model size from value."""
        self._size_input_model = float(value)

    @property
    def size_optimized_model(self) -> Optional[float]:
        """Get optimized model size."""
        return self._size_optimized_model

    @size_optimized_model.setter
    def size_optimized_model(self, value: str) -> None:
        """Set optimized  model size optimized from value."""
        self._size_optimized_model = float(value)

    @property
    def path_optimized_model(self) -> Optional[str]:
        """Get optimized model path."""
        return self._path_optimized_model

    @path_optimized_model.setter
    def path_optimized_model(self, value: str) -> None:
        """Set optimized model path."""
        self._path_optimized_model = value

    @property
    def optimization_time(self) -> Optional[float]:
        """Optimization time."""
        return self.optimization_time

    @optimization_time.setter
    def optimization_time(self, value: str) -> None:
        """Set optimization_time value."""
        self.optimization_time = float(value)



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""UX server utils module."""

from typing import Any


def float_representer(dumper: Any, value: float) -> Any:
    """Float representer for yaml dumper."""
    if value.is_integer():
        float_out = f"{int(value)}.0"
    else:
        float_out = f"{value}"
    return dumper.represent_scalar("tag:yaml.org,2002:float", float_out)



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Execute benchmark."""

import json
import os
from typing import Any, Dict, List

from neural_compressor.ux.components.benchmark import Benchmarks
from neural_compressor.ux.components.benchmark.benchmark import Benchmark
from neural_compressor.ux.utils.exceptions import ClientErrorException, InternalException
from neural_compressor.ux.utils.executor import Executor
from neural_compressor.ux.utils.logger import log
from neural_compressor.ux.utils.parser import BenchmarkParserFactory
from neural_compressor.ux.utils.templates.workdir import Workdir
from neural_compressor.ux.utils.utils import _load_json_as_dict
from neural_compressor.ux.utils.workload.workload import ExecutionMode, Workload
from neural_compressor.ux.web.communication import MessageQueue

mq = MessageQueue()


def execute_benchmark(data: Dict[str, Any]) -> None:
    """
    Execute benchmark setting correct workload status.

    Expected data:
    {
        "id": "configuration_id",
        "input_model": {
            "precision": "fp32",
            "path": "/localdisk/fp32.pb"
        },
        "optimized_model": {
            "precision": "int8",
            "path": "/localdisk/int8.pb"
        }
    }
    """
    request_id = str(data.get("id", ""))

    if not request_id:
        message = "Missing request_id."
        mq.post_error(
            "benchmark_finish",
            {"message": "Failure", "code": 404},
        )
        log.error(f"Problem during benchmark: {message}")
        return

    workdir = Workdir(request_id=request_id, overwrite=False)
    workdir.set_workload_status(request_id, "wip")

    try:
        execute_real_benchmark(
            request_id=request_id,
            workdir=workdir,
            data=data,
        )
    except Exception:
        workdir.set_workload_status(request_id, "error")
        mq.post_error(
            "benchmark_finish",
            {"message": "Failure", "code": 404, "id": request_id},
        )
        log.exception("Exception during benchmark")
        return

    workdir.set_workload_status(request_id, "success")


def execute_real_benchmark(
    request_id: str,
    workdir: Workdir,
    data: Dict[str, Any],
) -> None:
    """
    Execute benchmark.

    Expected data:
    {
        "id": "configuration_id",
        "input_model": {
            "precision": "fp32",
            "path": "/localdisk/fp32.pb"
        },
        "optimized_model": {
            "precision": "int8",
            "path": "/localdisk/int8.pb"
        }
    }
    """
    from neural_compressor.ux.utils.workload.workload import Workload

    input_model = data.get("input_model", None)
    input_model.update({"model_type": "input_model"})

    optimized_model = data.get("optimized_model", None)
    optimized_model.update({"model_type": "optimized_model"})

    if not (input_model and optimized_model):
        message = "Missing input or optimized model data."
        raise ClientErrorException(message)

    workload_path = workdir.workload_path
    workload_data = _load_json_as_dict(
        os.path.join(workload_path, "workload.json"),
    )

    workload = Workload(workload_data)

    response_data: Dict[str, Any] = {"id": request_id, "execution_details": {}}

    mq.post_success(
        "benchmark_start",
        {
            "message": "started",
            "id": request_id,
        },
    )

    models = [input_model, optimized_model]

    benchmark_count = 0
    benchmark_total = 0

    for model_info in models:
        model_precision = model_info.get("precision", None)
        benchmark_modes: List[str] = model_info.get("mode", [Benchmarks.PERF])
        if (
            ExecutionMode.ADVANCED == workload.execution_mode
            and "fp32" == workload.output_precision
            and Benchmarks.ACC not in benchmark_modes
        ):  # Accuracy information is provided only in tuning
            benchmark_modes.append(Benchmarks.ACC)
        model_info.update({"benchmark_modes": benchmark_modes})
        benchmark_total += len(benchmark_modes)

    for model_info in models:
        model_precision = model_info.get("precision", None)
        model_type = model_info.get("model_type", None)
        model_path = model_info.get("path", None)
        benchmark_modes = model_info.get("benchmark_modes", None)

        if not (model_precision and model_path and model_type and benchmark_modes):
            message = "Missing model precision, model path or model type."
            raise ClientErrorException(message)

        for benchmark_mode in benchmark_modes:
            benchmark_count += 1
            response_data = benchmark_model_and_respond_to_ui(
                response_data=response_data,
                workload=workload,
                workdir=workdir,
                model=model_type,
                model_path=model_path,
                model_precision=model_precision,
                benchmark_mode=benchmark_mode,
                benchmark_count=benchmark_count,
                benchmark_total=benchmark_total,
            )
        workdir.set_workload_status(request_id, "success")


def benchmark_model_and_respond_to_ui(
    response_data: dict,
    workload: Workload,
    workdir: Workdir,
    model: str,
    model_path: str,
    model_precision: str,
    benchmark_mode: str,
    benchmark_count: int,
    benchmark_total: int,
) -> dict:
    """Benchmark model and update UI."""
    response_data = benchmark_model(
        response_data=response_data,
        workload=workload,
        workdir=workdir,
        model=model,
        model_path=model_path,
        model_precision=model_precision,
        benchmark_mode=benchmark_mode,
        benchmark_count=benchmark_count,
        benchmark_total=benchmark_total,
    )
    mq.post_success("benchmark_finish", response_data)
    return response_data


def benchmark_model(
    response_data: dict,
    workload: Workload,
    workdir: Workdir,
    model: str,
    model_path: str,
    model_precision: str,
    benchmark_mode: str,
    benchmark_count: int,
    benchmark_total: int,
) -> dict:
    """Benchmark model and prepare response data."""
    request_id = response_data.get("id")

    benchmark: Benchmark = Benchmark(
        workload=workload,
        model_path=model_path,
        precision=model_precision,
        mode=benchmark_mode,
    )

    log_name = f"{model}_{benchmark_mode}_benchmark"

    executor = Executor(
        workload.workload_path,
        subject="benchmark",
        data={"id": request_id},
        send_response=False,
        log_name=log_name,
        additional_log_names=["output.txt"],
    )

    proc = executor.call(
        benchmark.command,
    )

    logs = [os.path.join(workload.workload_path, f"{log_name}.txt")]

    if not proc.is_ok:
        raise ClientErrorException("Benchmark failed during execution.")

    parser = BenchmarkParserFactory.get_parser(benchmark_mode, logs)
    metrics = parser.process()
    metric = {}
    execution_details: Dict[str, Any] = {}

    if benchmark_mode == Benchmarks.PERF:
        result_field = f"perf_throughput_{model}"
    elif benchmark_mode == Benchmarks.ACC:
        result_field = f"acc_{model}"
    else:
        raise InternalException(f"Benchmark mode {benchmark_mode} is not supported.")

    if isinstance(metrics, dict):
        metric = {result_field: metrics.get(result_field, "")}
        execution_details = response_data.get("execution_details", {})
        model_benchmark_details = execution_details.get(f"{model}_benchmark", {})
        model_benchmark_details.update(
            {
                benchmark_mode: benchmark.serialize(),
            },
        )

        response_data.update(
            {
                "progress": f"{benchmark_count}/{benchmark_total}",
                "current_step": benchmark_count,
                "number_of_steps": benchmark_total,
            },
        )
        response_data.update(metric)
        response_data["execution_details"].update(
            {f"{model}_benchmark": model_benchmark_details},
        )
    workdir.update_metrics(
        request_id=request_id,
        metric_data=metric,
    )
    workdir.update_execution_details(
        request_id=request_id,
        execution_details=execution_details,
    )
    log.debug(f"Parsed data is {json.dumps(response_data)}")
    mq.post_success("benchmark_progress", response_data)

    return response_data



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------
import os
import pickle
import struct
from pathlib import Path
from engine.converter.ops.op import OPERATORS, Tensor
import numpy as np
from collections import namedtuple
from collections import OrderedDict

import copy

class EngineQuantizer:
    def __init__(self, model, dataloader, iterations, q_config, op_types_to_quantize):
        self.model = model
        self.dataloader = dataloader
        self.iterations = iterations
        self.config = q_config
        self._tensors_min_max = {}
        self._quantize_op = op_types_to_quantize
        self._quantize_input_op = ["Quantize"]
        self._quantize_output_op = ["Softmax"]
        self._part_quantize_op = self._quantize_input_op + self._quantize_output_op
        self._no_quantize_op = ["Reshape", "Reorder"]
        
    def quantize_model(self):
        # 0. get min/max of tensor
        self._get_tensors_min_max()
        # 1. quant innerproduct weight and bias
        self._quantize_innerproduct()
        # 2. insert quantize before innerproduct
        self._insert_quantize_op()
        # 3. insert quant info to each tensor
        self._insert_quantize_info()
        # 4. int8 parttern fusion
        self._int_pattern_fusion()
        return self.model

    def _get_tensors_min_max(self):
        self._matmul_node_input_fp32()
        net = self.model.graph.dump_tensor()
        output_tensors = [content['input'] for content in \
            net['model']['operator'].values() if content['type'] == 'Output'][0]
        for output_tensor_name in output_tensors:
            self._tensors_min_max[output_tensor_name] = [float('inf'), float('-inf')]
        # get min/max of const tensors
        for node in self.model.nodes:
            weight_perm = node.attr['src1_perm'] if node.op_type == "InnerProduct" else None
            for input_tensor in node.input_tensors:
                if not input_tensor.source_op:
                    self._tensors_min_max[input_tensor.name] = [float('inf'), float('-inf')]
                    quantize_mode = 'per_tensor' if node.name not in self.config or \
                            self.config[node.name] =='fp32' else \
                            self.config[node.name]['weight']['granularity']
                    self._get_min_max(input_tensor, weight_perm, quantize_mode,
                                      self._tensors_min_max[input_tensor.name])

        self.model.graph.engine_init(net)
        for idx, (inputs, labels) in enumerate(self.dataloader):
            if idx > self.iterations:
                break
            else:
                # inference and put the tensors data to graph
                results = self.model.graph.inference(inputs)
                for tensor_name, tensor_data in results.items():
                    for node in self.model.nodes:
                        for input_tensor in node.input_tensors:
                            if tensor_name == input_tensor.name:
                                input_tensor.data = tensor_data
                                input_tensor.dtype = 'fp32'
                        for output_tensor in node.output_tensors:
                            if tensor_name == output_tensor.name:
                                output_tensor.data = tensor_data
                                output_tensor.dtype = 'fp32'
                
                # get min/max of output tensors
                for node in self.model.nodes:
                    for output_tensor in node.output_tensors:
                        if output_tensor.name in self._tensors_min_max and output_tensor.source_op:
                            quantize_mode = 'per_tensor' if node.name not in self.config or \
                                self.config[node.name] =='fp32' else\
                                self.config[node.name]['weight']['granularity']
                            self._get_min_max(output_tensor, None,
                                              quantize_mode,
                                              self._tensors_min_max[output_tensor.name])

    def _get_min_max(self, tensor, weight_perm, quantize_mode, tensor_min_max):
        tensor_name = tensor.name
        tensor_data = tensor.data
        tensor_min_history = tensor_min_max[0]
        tensor_max_history = tensor_min_max[1]
        # get min
        tensor_min_data = np.min(tensor_data)
        if not tensor.source_op and quantize_mode == "per_channel":
            if weight_perm == '0,1':
                tensor_min_data = np.min(tensor_data, axis=-1)
            else:
                tensor_min_data = np.min(tensor_data, axis=0)
        tensor_min_data = np.minimum(tensor_min_data, 0.).astype(np.float32)
        tensor_min_history = np.minimum(tensor_min_data, tensor_min_history).astype(np.float32)
        # get max
        tensor_max_data = np.max(tensor_data)
        if not tensor.source_op and quantize_mode == "per_channel":
            if weight_perm == '0,1':
                tensor_max_data = np.max(tensor_data, axis=-1)
            else:
                tensor_max_data = np.max(tensor_data, axis=0)
        tensor_max_data = np.maximum(tensor_max_data, 0.).astype(np.float32)
        tensor_max_history = np.maximum(tensor_max_data, tensor_max_history).astype(np.float32)
        self._tensors_min_max[tensor_name] = [np.array(tensor_min_history).astype(np.float32),
                                            np.array(tensor_max_history).astype(np.float32)]

    def _matmul_node_input_fp32(self):
        # if one of matmul pre nodes is fp32, both are fp32
        for node in self.model.nodes:
            if node.op_type == 'Matmul':
                is_fp32 = self._is_pre_matmul_fp32(node)
                if is_fp32:
                    self._another_pre_matmul_fp32(node)
                    self.config[node.name] ='fp32'

   # if pre node of matmul is in ordered to be fp32
    def _is_pre_matmul_fp32(self, node):
        pre_nodes_name = self.model.graph.get_pre_node_names(node.name)
        for pre_node_name in pre_nodes_name:
            pre_node_idx = self.model.get_node_id(pre_node_name)
            pre_node = self.model.nodes[pre_node_idx] 
            if pre_node.name in self.config and self.config[pre_node.name] =='fp32':
                return True
            elif pre_node.op_type in self._no_quantize_op:
                return self._is_pre_matmul_fp32(pre_node)
        return False

    # if another pre matmul quantized node is fp32
    def _another_pre_matmul_fp32(self, node):
        pre_nodes_name = self.model.graph.get_pre_node_names(node.name)
        for pre_node_name in pre_nodes_name:
            pre_node_idx = self.model.get_node_id(pre_node_name)
            pre_node = self.model.nodes[pre_node_idx]
            if pre_node.op_type in (self._quantize_op + self._part_quantize_op):
                self.config[pre_node.name] ='fp32'
            elif pre_node.op_type in self._no_quantize_op:
                self._another_pre_matmul_fp32(pre_node)

    def _insert_quantize_op(self):
        nodes = copy.deepcopy(self.model.nodes)
        for ip_node in nodes: 
            if ip_node.op_type == "InnerProduct" and self.config[ip_node.name] != 'fp32':
                ip_node_name = ip_node.name
                ip_node_idx = self.model.get_node_id(ip_node_name)
                ip_input_tensor = ip_node.input_tensors[0]
                pre_ip_node_name = ip_input_tensor.source_op[0]

                # pre ip_node
                pre_ip_node_idx = self.model.get_node_id(pre_ip_node_name)
                pre_ip_node = self.model.nodes[pre_ip_node_idx]
                ip_input_tensor = copy.deepcopy(ip_node.input_tensors[0])

                # 1. insert quantize op before innerproduct
                # change pre ip op's next op to quant op
                quant_node_name = ip_node_name + "_quant"
                # create quant output tensor 
                quant_output_tensor_name = ip_input_tensor.name + '_quant'
                if self.config[ip_node.name]['activation']['scheme'] == 'asym':
                    quant_output_tensor = Tensor(
                                            name = quant_output_tensor_name,
                                            source_op = [quant_node_name],
                                            dest_op = [ip_node.name],
                                            dtype='u8'
                                            )
                else:
                    quant_output_tensor = Tensor(
                                            name = quant_output_tensor_name,
                                            source_op = [quant_node_name],
                                            dest_op = [ip_node.name],
                                            dtype='s8'
                                            )                    
                self._tensors_min_max[quant_output_tensor_name] = \
                    self._tensors_min_max[ip_input_tensor.name]
                # create quant op
                quant_node = OPERATORS['QuantizeV2']()
                if self.config[ip_node.name]['activation']['scheme'] == 'asym':
                    quant_node.construct(name=quant_node_name, op_type='Quantize', 
                                        input_tensors=[ip_input_tensor],
                                        output_tensors=[quant_output_tensor],
                                        attr=OrderedDict({'output_dtype':'u8'}))
                else:
                    quant_node.construct(name=quant_node_name, op_type='Quantize', 
                                        input_tensors=[ip_input_tensor],
                                        output_tensors=[quant_output_tensor],
                                        attr=OrderedDict({'output_dtype':'s8'}))

                self.model.insert_nodes(ip_node_idx, [quant_node])
                # innerproduct input replace quantize output
                self.model.graph.change_node_input_tensors(ip_node_name, 0, quant_output_tensor)

        self._remove_duplicate_quantize_node()

    def _remove_duplicate_quantize_node(self):
        visit_tensors = []
        nodes = copy.deepcopy(self.model.nodes)
        for node in nodes:
            if node.op_type == "Quantize":
                input_name = node.input_tensors[0].name
                if input_name in visit_tensors:
                    self.model.remove_nodes([node.name])
                else:    
                    visit_tensors.append(input_name) 

    def _insert_quantize_info(self):
        for node in self.model.nodes:
            if (node.op_type in self._quantize_op and \
                self.config[node.name.split('_quant')[0]] != 'fp32') or \
                node.op_type in self._part_quantize_op:

                # insert input tensor min max
                node_type = node.op_type
                input_tensors = copy.deepcopy(node.input_tensors)
                input_size = len(input_tensors)

                # Don't quantize bias of inner_product and matmul
                input_size = 2 if input_size > 2 else input_size

                if node_type not in self._quantize_output_op:
                    for idx in range(input_size):
                        tensor = node.input_tensors[idx]
                        tensor_name = tensor.name
                        tensor_min_data = self._tensors_min_max[tensor_name][0]
                        tensor_min_shape = list(tensor_min_data.shape) \
                            if tensor_min_data.shape != () else [1]
                        tensor_min_name = tensor_name + '_min'
                        tensor_min = Tensor(name = tensor_min_name,
                                            dest_op = [node.name],
                                            data = tensor_min_data,
                                            shape = tensor_min_shape,
                                            dtype='fp32'
                                            )
                        node.input_tensors.append(tensor_min)

                        tensor_max_data = self._tensors_min_max[tensor_name][1]
                        tensor_max_shape = list(tensor_max_data.shape) \
                            if tensor_max_data.shape != () else [1]
                        tensor_max_name = tensor_name + '_max'
                        tensor_max = Tensor(name = tensor_max_name,
                                            dest_op = [node.name],
                                            data = tensor_max_data,
                                            shape = tensor_max_shape,
                                            dtype='fp32'
                                            )
                        node.input_tensors.append(tensor_max)

                # 2. insert output tensor min max
                if node_type not in self._quantize_input_op:
                    output_tensors = copy.deepcopy(node.output_tensors)
                    for idx in range(len(output_tensors)):
                        tensor = node.output_tensors[idx]
                        tensor_name = tensor.name
                        tensor_data = self._tensors_min_max[tensor_name]

                        tensor_min_data = self._tensors_min_max[tensor_name][0]
                        tensor_min_shape = list(tensor_min_data.shape) \
                            if tensor_min_data.shape != () else [1]
                        tensor_min_name = tensor_name + '_min'
                        tensor_min = Tensor(name = tensor_min_name,
                                            dest_op = [node.name],
                                            data = tensor_min_data,
                                            shape = tensor_min_shape,
                                            dtype='fp32'
                                            )
                        node.input_tensors.append(tensor_min)

                        tensor_max_data = self._tensors_min_max[tensor_name][1]
                        tensor_max_shape = list(tensor_max_data.shape) \
                            if tensor_max_data.shape != () else [1]
                        tensor_max_name = tensor_name + '_max'
                        tensor_max = Tensor(name = tensor_max_name,
                                            dest_op = [node.name],
                                            data = tensor_max_data,
                                            shape = tensor_max_shape,
                                            dtype='fp32'
                                            )
                        node.input_tensors.append(tensor_max)

    def _quantize_innerproduct(self):
        for node in self.model.nodes:
            if node.op_type == "InnerProduct" and \
                self.config[node.name.split('_quant')[0]] != 'fp32':

                input_tensor = node.input_tensors[0]
                weight_tensor = node.input_tensors[1]
                weight_perm = node.attr['src1_perm'] 
                
                self._quant_weight(weight_tensor, weight_perm)

                bias_tensor = node.input_tensors[2] if len(node.input_tensors) > 2 else None
                if (bias_tensor and not bias_tensor.source_op) or not bias_tensor:
                    self._quant_bias(node, input_tensor, weight_tensor, bias_tensor, weight_perm)

    def _get_scale_zp(self, tensor, dtype):
        tensor_name = tensor.name
        tensor_min_data = self._tensors_min_max[tensor_name][0]
        tensor_max_data = self._tensors_min_max[tensor_name][1]

        if (dtype=='u8'):
            max_sub_min = tensor_max_data - tensor_min_data
            scale = np.where(max_sub_min==0., 0., 255./max_sub_min)
            zero_point = -tensor_min_data * scale
        elif (dtype=='s8'):
            max_abs = np.maximum(np.fabs(tensor_min_data), np.fabs(tensor_max_data))
            scale = np.where(max_abs==0., 0., 127./max_abs)
            zero_point = 0.

        return scale, zero_point

    def _quant_weight(self, weight_tensor, weight_perm):
        weight_data = weight_tensor.data
        weight_scale , weight_zero_point = self._get_scale_zp(weight_tensor, 's8')

        if weight_perm=="0,1":
            weight_tensor.data = \
                (np.round((weight_tensor.data).T * weight_scale).astype(np.int8)).T
        else:
            weight_tensor.data = np.round(weight_tensor.data * weight_scale).astype(np.int8)
        weight_tensor.dtype = 's8'

    def _quant_bias(self, node, input_tensor, weight_tensor, bias_tensor, weight_perm):
        input_data_min = self._tensors_min_max[input_tensor.name][0]
        weight_s8 = weight_tensor.data
        if self.config[node.name]['activation']['scheme'] == 'asym':
            input_scale , input_zero_point = self._get_scale_zp(input_tensor, 'u8')
        else:
            input_scale , input_zero_point = self._get_scale_zp(input_tensor, 's8')
        weight_scale , weight_zero_point = self._get_scale_zp(weight_tensor, 's8')

        if self.config[node.name]['activation']['scheme'] == 'asym':
            if weight_perm == '0,1':
                bias_zero_point = input_scale * input_data_min * \
                    np.sum(weight_s8.astype(float), axis=-1)
            else:
                bias_zero_point = input_scale * input_data_min * \
                    np.sum(weight_s8.astype(float), axis=0)
            bias_s32 = bias_tensor.data * input_scale * weight_scale if bias_tensor else 0.
            bias_s32 = np.round(bias_s32 + bias_zero_point).astype(np.int32)
            if not bias_tensor:
                bias_tensor_name = weight_tensor.name + "_bias"
                bias_shape = list(bias_zero_point.shape) \
                    if bias_zero_point.shape != () else [1]
                bias_tensor = Tensor(name = bias_tensor_name,
                                dest_op = [node.name],
                                data = bias_s32,
                                shape = bias_shape,
                                dtype='s32'
                                )
                node.input_tensors.append(bias_tensor)
            else:
                bias_tensor.data = bias_s32
                bias_tensor.dtype = 's32'
        else:
            if bias_tensor:
                bias_s32 = np.round(bias_tensor.data * input_scale * weight_scale).astype(np.int32)
                bias_tensor.data = bias_s32
                bias_tensor.dtype = 's32'

    def _int_pattern_fusion(self):
        self._quantize_node_out_int8()
        self._quantize_fusion()

    # quant op output int8
    def _quantize_node_out_int8(self):
        for node in self.model.nodes:
            if (node.op_type in self._quantize_op and \
                    self.config[node.name.split('_quant')[0]] != 'fp32') or \
                    node.op_type in self._quantize_output_op:
                is_output_int8 = True 
                is_output_int8 = self._is_output_int8(node, is_output_int8)
                if is_output_int8:
                    if node.op_type == "Softmax":
                        node.output_tensors[0].dtype = 'u8'
                        node.attr['output_dtype'] = 'u8'
                    else:
                        node.output_tensors[0].dtype = 's8'
                        node.attr['output_dtype'] = 's8'

    # if quant op can output int8
    def _is_output_int8(self, node, is_output_int8):
        for output in node.output_tensors:
            for next_node_name in output.dest_op:
                next_node_idx = self.model.get_node_id(next_node_name)
                next_node = self.model.nodes[next_node_idx]
                if next_node.op_type in self._quantize_op and \
                        self.config[next_node.name.split('_quant')[0]] != 'fp32':
                    return is_output_int8 & True
                elif next_node.op_type in self._no_quantize_op:
                    return self._is_output_int8(next_node, is_output_int8)
                else:
                    return is_output_int8 & False

    # fuse quantize_output_op with quantize
    def _quantize_fusion(self):
        for node in self.model.nodes:
            if (node.op_type in self._quantize_op and \
                    self.config[node.name.split('_quant')[0]] != 'fp32') or \
                    node.op_type in self._quantize_output_op:
                is_quantize_fusion = True
                is_quantize_fusion, pre_quantize_node, quantize_node = self._is_quantize_fusion(
                    node, is_quantize_fusion)
                if is_quantize_fusion:
                    # if node.op_type != 'InnerProduct':
                    self.model.remove_nodes([quantize_node.name])
                    pre_quantize_node.output_tensors[0] = quantize_node.output_tensors[0]
                    if self.config[node.name]['activation']['scheme'] == 'asym':
                        node.attr['output_dtype'] = 'u8'
                    else:
                        node.attr['output_dtype'] = 's8'

    # if quantize_output_op can fuse quantize
    def _is_quantize_fusion(self, node, is_qunatize_fusion):
        for output in node.output_tensors:
            for next_node_name in output.dest_op:
                next_node_idx = self.model.get_node_id(next_node_name)
                next_node = self.model.nodes[next_node_idx]
                if next_node.op_type == "Quantize":
                    is_qunatize_fusion &= True
                    return is_qunatize_fusion, node, next_node
                elif next_node.op_type in self._no_quantize_op:
                    return self._is_quantize_fusion(next_node, is_qunatize_fusion)
                else:
                    is_qunatize_fusion &= False
                    return is_qunatize_fusion & False, None, None

        is_qunatize_fusion &= False
        return is_qunatize_fusion, None, None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .pattern import supported_patterns, PATTERNS
from neural_compressor.utils import logger

EXECUTOR_TYPE = {
    "MatMulWithBias": "InnerProduct",
    "MatMulWithBiasAdd": "InnerProduct",
    "MatMulWithBiasGelu": "InnerProduct",
    "MatMulWithBiasTanh": "InnerProduct",
    "MatMulWithBiasRelu": "InnerProduct",
    "MatMul": "InnerProduct",
    "QuantizedMatMulWithBiasAndDequantize": "InnerProduct",
    "TransposeBatchMatMul": "Matmul",
    "BatchMatMul": "Matmul",
    "BatchMatMulV2": "Matmul",
    "Add": "BinaryAdd",
    "AddV2": "BinaryAdd",
    "AddWithAdd": "BinaryAdd",
    "QLinearAdd": "BinaryAdd",
    "Transpose": "Reorder",
    "GatherV2": "Gather",
    "ExpandDimsToReshape": "Reshape",
    "QuantizeV2": "Quantize",
    "QuantizeLinear": "Quantize",
    "OneHot": "Onehot",
    "LayerNormalization": "LayerNorm",
    "FusedGemm": "InnerProduct",
    "_QuantizedFusedMatMulAndDequantize": "InnerProduct",
    "_FusedMatMul": "InnerProduct",
    "_MklLayerNorm": "LayerNorm",
}


class SubGraphMatcher(object):
    def __call__(self, model):
        patterns_switch = {
            'LayerNorm': True,
            'TransposeBatchMatMul': True,
            'MatMulWithBiasGelu': True,
            'MatMulWithBiasAdd': True,
            'MatMulWithBiasTanh': True,
        }
        logger.info('Start to implement Sub-Graph matching and replacing...')
        for pattern in supported_patterns:
            if pattern in PATTERNS:
                if pattern in patterns_switch.keys() and not patterns_switch[pattern]:
                    continue
                else:
                    p_fusion = PATTERNS[pattern]()
                    model = p_fusion(model)

        rm_node_names = []
        rm_op_type = ['Identity']
        for i in range(len(model.nodes)):
            node = model.nodes[i]
            if node.op_type in rm_op_type:
                rm_node_names.append(node.name)
            else:
                if node.op_type in EXECUTOR_TYPE.keys():
                    op_type = EXECUTOR_TYPE[node.op_type]
                    model.nodes[i].op_type = op_type
        model.remove_nodes(rm_node_names)
        logger.info('Sub-Graph match and replace done...')

        return model



# -*- coding: utf-8 -*-
# Copyright (c) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Configuration evaluation module."""

from typing import Any, Dict, List, Optional, Union

from neural_compressor.ux.utils.hw_info import HWInfo
from neural_compressor.ux.utils.json_serializer import JsonSerializer
from neural_compressor.ux.utils.workload.dataloader import Dataloader


class Metric(JsonSerializer):
    """Configuration Metric class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Metric class."""
        super().__init__()
        self._name: str = data.get("name", None)
        self._param: Optional[Union[int, bool]] = data.get("param", None)

        if len(data) == 1 and not (self.name and self.param):
            self.name = list(data.keys())[0]
            self.param = data.get(self.name, None)

    @property
    def name(self) -> str:
        """Get metric name."""
        return self._name

    @name.setter
    def name(self, value: str) -> None:
        """Set metric name."""
        self._name = value

    @property
    def param(self) -> Optional[Union[int, bool]]:
        """Get metric param."""
        return self._param

    @param.setter
    def param(self, value: Union[int, bool]) -> None:
        """Set metric param."""
        if not (isinstance(value, str) and value == ""):
            self._param = value

    def serialize(
        self,
        serialization_type: str = "default",
    ) -> Union[Dict[str, Any], List[Dict[str, Any]]]:
        """Return metric dict for config."""
        if self.param and isinstance(self.param, dict):
            return {self.name: self.param}
        if self.name in ["MSE", "RMSE", "MAE"]:
            return {self.name: {"compare_label": self.param}}
        if self.name in ["COCOmAP"] and self.param:
            return {self.name: {"anno_path": self.param}}
        if self.name:
            return {self.name: self.param}
        return {}


class Configs(JsonSerializer):
    """Configuration Configs class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Configs class."""
        super().__init__()

        self.cores_per_instance: int = data.get("cores_per_instance", 4)
        self.num_of_instance: int = data.get(
            "num_of_instance",
            HWInfo().cores // self.cores_per_instance,
        )
        self.inter_num_of_threads: int = data.get("inter_num_of_threads", None)
        self.intra_num_of_threads: int = data.get("intra_num_of_threads", None)
        self.kmp_blocktime: int = data.get("kmp_blocktime", 1)


class PostprocessSchema(JsonSerializer):
    """Configuration PostprocessSchema class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration PostprocessSchema class."""
        super().__init__()
        self.LabelShift = data.get("LabelShift", None)  # [Optional] >0
        self.SquadV1 = data.get("SquadV1", None)
        self.Collect = data.get("Collect", None)


class Postprocess(JsonSerializer):
    """Configuration Postprocess class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Postprocess class."""
        super().__init__()
        self.transform = None
        if isinstance(data.get("transform"), dict):
            self.transform = PostprocessSchema(data.get("transform", {}))


class Accuracy(JsonSerializer):
    """Configuration Accuracy class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Accuracy class."""
        super().__init__()
        self.metric = None
        if isinstance(data.get("metric"), dict):
            self.metric = Metric(data.get("metric", {}))

        self.configs = None
        if isinstance(data.get("configs"), dict):
            self.configs = Configs(data.get("configs", {}))

        self.dataloader = None
        if isinstance(data.get("dataloader"), dict):
            self.dataloader = Dataloader(data.get("dataloader", {}))

        self.postprocess = None
        if isinstance(data.get("postprocess"), dict):
            self.postprocess = Postprocess(data.get("postprocess", {}))


class Performance(JsonSerializer):
    """Configuration Performance class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Performance class."""
        super().__init__()
        self.warmup: int = data.get("warmup", 5)

        self.iteration: int = data.get("iteration", -1)

        self.configs: Configs = Configs(data.get("configs", {}))

        self.dataloader: Optional[Dataloader] = None
        if isinstance(data.get("dataloader"), dict):
            self.dataloader = Dataloader(data.get("dataloader", {}))

        self.postprocess: Optional[Postprocess] = None
        if isinstance(data.get("postprocess"), dict):
            self.postprocess = Postprocess(data.get("postprocess", {}))


class Evaluation(JsonSerializer):
    """Configuration Evaluation class."""

    def __init__(self, data: Dict[str, Any] = {}):
        """Initialize Configuration Evaluation class."""
        super().__init__()
        self.accuracy: Optional[Accuracy] = None
        if isinstance(data.get("accuracy"), dict):
            self.accuracy = Accuracy(data.get("accuracy", {}))

        self.performance = None
        if isinstance(data.get("performance"), dict):
            self.performance = Performance(data.get("performance", {}))
